# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for post-processing of raw results data.

"""
from __future__ import annotations

import warnings
from math import isclose
from typing import Any, Union


def normalize_batch_bit_lengths(measurements: list[dict[str, int]]) -> list[dict[str, int]]:
    """
    Normalizes the bit lengths of binary keys in measurement count dictionaries
    to ensure uniformity across all keys.

    Args:
        measurements (list[dict[str, int]]): A list of dictionaries where each dictionary
            contains binary string keys and integer values.

    Returns:
        list[dict[str, int]]: A new list of dictionaries with uniformly lengthened binary keys.
    """
    if len(measurements) == 0:
        return measurements

    first_len = None
    all_keys_same_length = True
    max_bit_length = 0

    for counts in measurements:
        for key in counts.keys():
            key_len = len(key)
            if first_len is None:
                first_len = key_len
            if key_len != first_len:
                all_keys_same_length = False
            max_bit_length = max(max_bit_length, key_len)

    if all_keys_same_length:
        return measurements

    normalized_counts_list = []
    for counts in measurements:
        normalized_counts = {}
        for key, value in counts.items():
            normalized_key = key.zfill(max_bit_length)
            normalized_counts[normalized_key] = value
        normalized_counts_list.append(normalized_counts)

    return normalized_counts_list


def normalize_bit_lengths(measurement: dict[str, int]) -> dict[str, int]:
    """
    Normalizes the bit lengths of binary keys in a single measurement count dictionary
        to ensure uniformity across all keys.

    Args:
        measurement (dict[str, int]): A dictionary with binary string keys and integer values.

    Returns:
        dict[str, int]: A dictionary with uniformly lengthened binary keys.
    """
    normalized_list = normalize_batch_bit_lengths([measurement])

    return normalized_list[0] if normalized_list else measurement


def format_data(
    data: dict[Any, Union[int, float]],
    include_zero_values: bool = False,
    decimal: bool = False,
) -> dict[Any, Union[int, float]]:
    """
    Formats and sorts a counts dictionary with binary or integer keys.

    Args:
        data (dict[Any, Union[int, float]]): Dictionary with keys as binary strings
            (e.g., '0 1') or integers (e.g., 3), and values as counts.
        include_zero_values (bool, optional): Include missing states with zero counts.
            Defaults to False.
        decimal (bool, optional): Convert binary keys to decimal integers. Defaults to False.

    Returns:
        dict[Any, Union[int, float]]: Sorted dictionary with formatted keys.

    Raises:
        ValueError: If keys are neither valid binary strings nor integers.

    Examples:
        Basic formatting with binary keys:
        >>> data = {'1 1': 13, '0 0': 46, '1 0': 79}
        >>> format_data(data)
        {'00': 46, '10': 79, '11': 13}

        Include zero values:
        >>> format_data(data,include_zero_values=True)
        {'00': 46, '01': 0, '10': 79, '11': 13}

        Convert binary keys to decimal:
        >>> format_data(data,decimal=True)
        {0: 46, 1: 0, 2: 79, 3: 13}
    """
    if not data:
        return data

    input_is_dec = False
    input_is_bin = False

    if all(isinstance(key, str) for key in data.keys()):
        if all(
            (key.startswith("0b") and set(key[2:]).issubset({"0", "1", " "}))
            or set(key).issubset({"0", "1", " "})
            for key in data.keys()
        ):
            key_str_data = {
                (key.replace(" ", "")[2:] if key.startswith("0b") else key.replace(" ", "")): value
                for key, value in data.items()
            }
            normalized_data = normalize_bit_lengths(key_str_data)
            num_bits = max(len(key) for key in normalized_data)
            all_keys = [format(i, f"0{num_bits}b") for i in range(2**num_bits)]
            data = {key: normalized_data.get(key, 0) for key in all_keys}
            input_is_bin = True
        elif all(key.isdigit() for key in data.keys()):
            data = {int(key): value for key, value in data.items()}
            input_is_dec = True

    if decimal:
        if input_is_bin:
            data = {int(key, 2): value for key, value in data.items()}
        elif input_is_dec:
            data = {int(key): value for key, value in data.items()}

        if include_zero_values:
            max_key = max(data)
            for i in range(max_key + 1):
                if i not in data:
                    data[i] = 0

    elif input_is_dec or all(isinstance(key, int) for key in data.keys()):
        data = {bin(key): value for key, value in data.items()}
        return format_data(data, include_zero_values=include_zero_values, decimal=False)

    if not include_zero_values:
        data = {key: value for key, value in data.items() if value != 0}

    return dict(sorted(data.items()))


def normalize_data(
    data: Union[dict[str, int], list[dict[str, int]]],
    include_zero_values: bool = False,
    decimal: bool = False,
) -> Union[dict[str, int], list[dict[str, int]]]:
    """Returns the sorted histogram data of the run"""
    if isinstance(data, dict):
        return format_data(data, include_zero_values=include_zero_values, decimal=decimal)

    batch_data = [
        format_data(datum, include_zero_values=include_zero_values, decimal=decimal)
        for datum in data
    ]

    if decimal:
        return batch_data

    return normalize_batch_bit_lengths(batch_data)


def _counts_to_probabilities(counts: dict[str, int]) -> dict[str, float]:
    """
    Convert histogram counts to probabilities.

    Args:
        counts (dict[str, int]): A dictionary with measurement outcomes as keys
            and their counts as values.

    Returns:
        dict[str, float]: A dictionary with measurement outcomes as keys and their
            probabilities as values.
    """
    total_counts = sum(counts.values())
    measurement_probabilities = {outcome: count / total_counts for outcome, count in counts.items()}
    return measurement_probabilities


def counts_to_probabilities(
    counts: Union[dict[Any, int], list[dict[Any, int]]],
) -> Union[dict[Any, float], list[dict[Any, float]]]:
    """Calculate and return the probabilities of each measurement result."""
    if isinstance(counts, dict):
        return _counts_to_probabilities(counts)

    return [_counts_to_probabilities(count) for count in counts]


def normalize_tuples(measurements: list[list[tuple[int, ...]]]) -> list[list[tuple[int, ...]]]:
    """
    Normalizes lists of tuples in a list to have the same tuple length across all entries
    by padding shorter tuples with zeros on the left.

    Args:
        measurements (list[list[tuple[int, ...]]]): A list of lists containing tuples
            with integer elements.

    Returns:
        list[list[tuple[int, ...]]]: A new list where each sublist's tuples have normalized
            lengths, preserving the binary significance of the numbers.
    """
    max_tuple_length = max(len(tup) for sublist in measurements for tup in sublist)

    normalized_measurements = []
    for sublist in measurements:
        normalized_sublist = []
        for tup in sublist:
            current_tuple = tuple(tup) if isinstance(tup, list) else tup
            padded_tuple = (0,) * (max_tuple_length - len(current_tuple)) + current_tuple
            normalized_sublist.append(padded_tuple)
        normalized_measurements.append(normalized_sublist)

    return normalized_measurements


def distribute_counts(probs: dict[Any, float], shots: int) -> dict[Any, int]:
    """
    Adjusts probabilistic counts to ensure the total equals the number of shots.

    Args:
        probs (dict[Any, float]): A dictionary mapping states to their probabilities.
        shots (int): The total number of shots to distribute among the states.

    Returns:
        dict[Any, int]: A dictionary mapping states to their adjusted counts such that
            the sum of counts equals `shots`.

    Raises:
        ValueError: If probabilities do not sum to 1, are not in the range [0, 1],
            or if shots is negative.

    Example:
        >>> probs = {0: 0.86, 1: 0.14}
        >>> shots = 10
        >>> distribute_counts(probs, shots)
        {0: 9, 1: 1}
    """
    if not isclose(sum(probs.values()), 1.0, rel_tol=1e-7):
        warnings.warn("Probabilities do not sum to 1.")

    if not all(0 <= prob <= 1 for prob in probs.values()):
        raise ValueError("Probabilities must be between 0 and 1.")

    if shots < 0:
        raise ValueError("Number of shots must be non-negative.")

    counts = {state: round(prob * shots) for state, prob in probs.items()}

    diff = shots - sum(counts.values())

    for state in sorted(counts.keys(), key=lambda k: -probs[k]):
        if diff == 0:
            break
        counts[state] += 1 if diff > 0 else -1
        diff -= 1 if diff > 0 else -1

    return counts



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for converting Braket circuits to/from OpenQASM

"""
import textwrap

import numpy as np
import qiskit
from braket.circuits import Circuit

from qbraid.interface import circuits_allclose
from qbraid.transpiler.conversions.braket import braket_to_qasm3
from qbraid.transpiler.conversions.qasm3 import qasm3_to_braket
from qbraid.transpiler.conversions.qiskit import qiskit_to_qasm3


def test_braket_to_qasm3_bell_circuit():
    """Test converting braket bell circuit to OpenQASM 3.0 string"""
    qasm_expected = """
    OPENQASM 3.0;
    bit[2] b;
    qubit[2] q;
    h q[0];
    cnot q[0], q[1];
    b[0] = measure q[0];
    b[1] = measure q[1];
    """

    qasm_expected_2 = """
    OPENQASM 3.0;
    bit[2] __bits__;
    qubit[2] __qubits__;
    h __qubits__[0];
    cnot __qubits__[0], __qubits__[1];
    __bits__[0] = measure __qubits__[0];
    __bits__[1] = measure __qubits__[1];
    """
    bell = Circuit().h(0).cnot(0, 1).measure(0).measure(1)
    qasm_out = braket_to_qasm3(bell).strip()
    qasm_expected = textwrap.dedent(qasm_expected).strip()
    qasm_expected_2 = textwrap.dedent(qasm_expected_2).strip()
    assert qasm_out in [qasm_expected, qasm_expected_2]


def test_braket_to_qasm3_bell_no_measurements():
    """Test converting braket circuit with no measurements to OpenQASM 3.0 string"""
    qasm_expected = """
    OPENQASM 3.0;
    bit[2] b;
    qubit[2] q;
    h q[0];
    cnot q[0], q[1];
    """

    qasm_expected_2 = """
    OPENQASM 3.0;
    bit[2] __bits__;
    qubit[2] __qubits__;
    h __qubits__[0];
    cnot __qubits__[0], __qubits__[1];
    """
    bell = Circuit().h(0).cnot(0, 1)
    qasm_out: str = braket_to_qasm3(bell).strip()
    qasm_expected = textwrap.dedent(qasm_expected).strip()
    qasm_expected_2 = textwrap.dedent(qasm_expected_2).strip()
    assert qasm_out in [qasm_expected, qasm_expected_2]


def test_braket_from_qasm3():
    """Test converting OpenQASM 3 string to braket circuit"""
    qasm = """
    OPENQASM 3.0;
    bit[2] b;
    qubit[2] q;
    rx(0.15) q[0];
    rx(0.3) q[1];
    """
    qasm = textwrap.dedent(qasm).strip()
    circuit_expected = Circuit().rx(0, 0.15).rx(1, 0.3)
    assert circuit_expected == qasm3_to_braket(qasm)


def test_qiskit_to_qasm3_to_braket():
    """Test converting Qiskit circuit to Braket via OpenQASM 3.0 for mapped gate defs"""
    qc = qiskit.QuantumCircuit(4)
    qc.cx(0, 1)
    qc.s(0)
    qc.sdg(1)
    qc.t(2)
    qc.tdg(3)
    qc.sx(0)
    qc.sxdg(1)
    qc.p(np.pi / 8, 3)
    qc.cp(np.pi / 4, 2, 3)

    qasm3_str = qiskit_to_qasm3(qc)
    circuit = qasm3_to_braket(qasm3_str)
    assert circuits_allclose(qc, circuit)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining QiskitCircuit Class

"""
from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING

import qiskit
from packaging import version
from qiskit.circuit import Qubit
from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.quantum_info import Operator
from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager

from qbraid.programs.exceptions import ProgramTypeError

from ._model import GateModelProgram

if TYPE_CHECKING:
    import numpy as np


class QiskitCircuit(GateModelProgram):
    """Wrapper class for ``qiskit.QuantumCircuit`` objects"""

    def __init__(self, program: qiskit.QuantumCircuit):
        super().__init__(program)
        if not isinstance(program, qiskit.QuantumCircuit):
            raise ProgramTypeError(
                message=f"Expected 'qiskit.QuantumCircuit' object, got '{type(program)}'."
            )

    @property
    def qubits(self) -> list[Qubit]:
        """Return the qubits acted upon by the operations in this circuit"""
        return self.program.qubits

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits in the circuit."""
        return self.program.num_qubits

    @property
    def num_clbits(self) -> int:
        """Return the number of classical bits in the circuit."""
        return self.program.num_clbits

    @property
    def depth(self) -> int:
        """Return the circuit depth (i.e., length of critical path)."""
        return self.program.depth()

    def _unitary(self) -> np.ndarray:
        """Calculate unitary of circuit. Removes measurement gates to
        perform calculation if necessary."""
        circuit = self.program.copy()
        circuit.remove_final_measurements()
        return Operator(circuit).data

    def remove_idle_qubits(self) -> None:
        """Checks whether the circuit uses contiguous qubits/indices,
        and if not, reduces dimension accordingly."""
        circuit = self.program.copy()

        dag = circuit_to_dag(circuit)

        idle_wires = list(dag.idle_wires())
        if version.parse(qiskit.__version__) >= version.parse("1.3"):
            idle_qubit_wires = []
            idle_clbit_wires = []

            for wire in idle_wires:
                if isinstance(wire, Qubit):
                    idle_qubit_wires.append(wire)
                else:
                    idle_clbit_wires.append(wire)

            dag.remove_qubits(*idle_qubit_wires)
            dag.remove_clbits(*idle_clbit_wires)
        else:  # pragma: no cover
            for w in idle_wires:
                dag._remove_idle_wire(w)
                try:
                    dag.qubits.remove(w)
                except ValueError:
                    pass

            dag.qregs = OrderedDict()

        self._program = dag_to_circuit(dag)

    def reverse_qubit_order(self) -> None:
        """Reverse the order of the qubits in the circuit."""
        circuit = self.program.copy()
        reversed_circuit = circuit.reverse_bits()
        self._program = reversed_circuit

    def transform(self, device) -> None:
        """Transform program to according to device target profile."""
        if getattr(device.profile, "local", False) is True:
            self.remove_idle_qubits()

        pm_option = device._options.get("pass_manager")
        pm = pm_option or generate_preset_pass_manager(backend=device._backend)
        isa_circuit = pm.run(self._program)
        self._program = isa_circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for retrieving and post-processing QuEra QASM simulator results.

"""
import json
from pathlib import Path
from typing import Any

import pandas as pd
import pytest
from qbraid_core.services.quantum.exceptions import QuantumServiceRequestError

from qbraid.runtime.native import QbraidJob
from qbraid.runtime.native.result import QuEraQasmSimulatorResultData
from qbraid.runtime.result import Result

try:
    from flair_visual.animation.runtime.qpustate import AnimateQPUState
    from flair_visual.simulation_result import QuEraSimulationResult

    flair_visual_installed = True
except ImportError:
    flair_visual_installed = False


MOCK_JOB_ID = "quera_qasm_simulator-jovyan-qjob-0123456789"


class MockClient:
    """Mock qbraid_core.services.quantum.QuantumClient."""

    def _load_json(self, file_name: str) -> dict[str, Any]:
        """Helper method to load JSON data from a file."""
        file_path = Path(__file__).parent / file_name
        with open(file_path, "r", encoding="utf-8") as json_file:
            return json.load(json_file)

    def _get_data_by_job_id(self, file_name: str, job_id: str) -> dict[str, Any]:
        """Helper method to fetch data by job ID from a JSON file."""
        data = self._load_json(file_name)
        if data.get("qbraidJobId") == job_id:
            return data
        raise QuantumServiceRequestError("No jobs found matching the given criteria")

    def get_job(self, job_id: str) -> dict[str, Any]:
        """Mock method to return the metadata corresponding to the specified quantum job."""
        return self._get_data_by_job_id("quera_simulation_data.json", job_id)

    # pylint: disable-next=unused-argument
    def get_job_results(self, job_id: str, **kwargs) -> dict[str, Any]:
        """Mock method to return the results of the quantum job with the given qBraid ID."""
        return self._get_data_by_job_id("quera_simulation_result.json", job_id)


@pytest.fixture
def mock_client() -> MockClient:
    """Fixture for a mock QuantumClient."""
    return MockClient()


@pytest.fixture
def mock_job(mock_client):
    """Fixture for a mock QbraidJob."""
    return QbraidJob(MOCK_JOB_ID, client=mock_client)


@pytest.mark.skipif(not flair_visual_installed, reason="flair-visual is not installed.")
def test_quera_qasm_simulator_job_result(mock_job: QbraidJob):
    """Test processing QuEra QASM simulator job results and result data."""
    result = mock_job.result()
    assert isinstance(result, Result)
    assert isinstance(result.data, QuEraQasmSimulatorResultData)

    assert result.data.backend == "cirq-gpu"
    assert result.data.flair_visual_version == "0.5.3"
    assert isinstance(result.data.quera_simulation_result, QuEraSimulationResult)
    assert isinstance(result.data.get_qpu_state(), AnimateQPUState)


@pytest.mark.skipif(not flair_visual_installed, reason="flair-visual is not installed.")
def test_quera_sim_data_get_logs_as_dataframe(mock_job: QbraidJob):
    """Test that get_logs returns the logs as a pandas DataFrame."""
    result = mock_job.result()
    result_data: QuEraQasmSimulatorResultData = result.data
    logs = result_data.get_logs()

    assert isinstance(logs, pd.DataFrame)
    assert logs.shape == (3, 5)  # 3 rows (3 log entries) and 5 columns
    assert list(logs.columns) == ["atom_id", "block_id", "action_type", "time", "duration"]


@pytest.mark.skipif(not flair_visual_installed, reason="flair-visual is not installed.")
def test_quera_sim_data_quera_simulator_result_value_error():
    """Test that get_qpu_state raises ValueError if quera_simulation_result is None."""
    result_data = QuEraQasmSimulatorResultData(backend="cirq", quera_simulation_result=None)

    with pytest.raises(ValueError, match="The simulation result is not available."):
        _ = result_data.quera_simulation_result



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for configuring IBM provider credentials and authentication.

"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING, Optional

import qiskit
from qiskit_ibm_runtime import QiskitRuntimeService
from qiskit_ibm_runtime.accounts import ChannelType

from qbraid._caching import cached_method
from qbraid.programs import ExperimentType, ProgramSpec
from qbraid.runtime.profile import TargetProfile
from qbraid.runtime.provider import QuantumProvider

from .device import QiskitBackend

if TYPE_CHECKING:
    import qiskit_ibm_runtime

    import qbraid.runtime.ibm


class QiskitRuntimeProvider(QuantumProvider):
    """
    This class is responsible for managing the interactions and
    authentications with the IBM Quantum services.

    Attributes:
        runtime_service (qiskit_ibm_runtime.QiskitRuntimeService): IBM Quantum runtime service.
    """

    def __init__(
        self,
        token: Optional[str] = None,
        instance: Optional[str] = None,
        channel: Optional[ChannelType] = None,
        **kwargs,
    ):
        """
        Initializes the QiskitRuntimeProvider object with IBM Quantum credentials.

        Args:
            token (str, optional): IBM Cloud API key.
            instance (str, optional): The service instance to use.
                This is the Cloud Resource Name (CRN) or the service name. If set, it will define
                a default instance for service instantiation. If not set, the service will fetch
                all instances accessible within the account.
            channel (ChannelType, optional): ``ibm_cloud``, ``ibm_quantum_platform`` or ``local``.
                If ``local``, uses testing mode and primitive queries will run on local simulator.
        """
        self.token = token or os.getenv("QISKIT_IBM_TOKEN")
        self.instance = instance or os.getenv("QISKIT_IBM_INSTANCE")
        self.channel = channel or os.getenv("QISKIT_IBM_CHANNEL", "ibm_quantum_platform")
        self._runtime_service = QiskitRuntimeService(
            channel=self.channel, token=self.token, instance=self.instance, **kwargs
        )

    @property
    def runtime_service(self) -> qiskit_ibm_runtime.QiskitRuntimeService:
        """Returns the IBM Quantum runtime service."""
        return self._runtime_service

    def save_config(
        self,
        token: Optional[str] = None,
        instance: Optional[str] = None,
        channel: Optional[ChannelType] = None,
        overwrite: bool = True,
        **kwargs,
    ) -> None:
        """Saves IBM runtime service account to disk for future use."""
        token = token or self.token
        instance = instance or self.instance
        channel = channel or self.channel
        QiskitRuntimeService.save_account(
            token=token, instance=instance, channel=channel, overwrite=overwrite, **kwargs
        )

    def _build_runtime_profile(
        self, backend: qiskit_ibm_runtime.IBMBackend, program_spec: Optional[ProgramSpec] = None
    ) -> TargetProfile:
        """Builds a runtime profile from a backend."""
        program_spec = program_spec or ProgramSpec(qiskit.QuantumCircuit)
        config = backend.configuration()

        local = config.local
        simulator = config.local or config.simulator

        return TargetProfile(
            device_id=backend.name,
            simulator=simulator,
            local=local,
            num_qubits=config.n_qubits,
            program_spec=program_spec,
            instance=backend._instance,
            max_shots=config.max_shots,
            provider_name="IBM",
            experiment_type=ExperimentType.GATE_MODEL,
            basis_gates=config.basis_gates,
        )

    @cached_method
    def get_devices(self, operational=True, **kwargs) -> list[qbraid.runtime.ibm.QiskitBackend]:
        """Returns the IBM Quantum provider backends."""

        backends = self.runtime_service.backends(operational=operational, **kwargs)
        program_spec = ProgramSpec(qiskit.QuantumCircuit)
        return [
            QiskitBackend(
                profile=self._build_runtime_profile(backend, program_spec=program_spec),
                service=self.runtime_service,
            )
            for backend in backends
        ]

    @cached_method
    def get_device(
        self, device_id: str, instance: Optional[str] = None
    ) -> qbraid.runtime.ibm.QiskitBackend:
        """Returns the IBM Quantum provider backends."""
        backend = self.runtime_service.backend(device_id, instance=instance)
        return QiskitBackend(
            profile=self._build_runtime_profile(backend), service=self.runtime_service
        )

    def least_busy(
        self, simulator=False, operational=True, **kwargs
    ) -> qbraid.runtime.ibm.QiskitBackend:
        """Return the least busy IBMQ QPU."""
        backend = self.runtime_service.least_busy(
            simulator=simulator, operational=operational, **kwargs
        )
        return QiskitBackend(
            profile=self._build_runtime_profile(backend), service=self.runtime_service
        )

    def __hash__(self):
        if not hasattr(self, "_hash"):
            object.__setattr__(self, "_hash", hash((self.token, self.channel)))
        return self._hash  # pylint: disable=no-member



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for submissions to AWS SV1, DM1, TN1 via qBraid native runtime.

"""

from qbraid.runtime.native.provider import _serialize_program


def test_braket_circuit_to_json(braket_circuit, qasm3_circuit):
    """Test conversion of Braket circuit to JSON."""
    qasm_json = _serialize_program(braket_circuit)
    assert len(qasm_json) == 1
    assert qasm_json.keys() == {"openQasm"}
    assert qasm_json["openQasm"].strip() == qasm3_circuit.strip()



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qasm2 cirq_custom gates

"""

import numpy as np
from cirq import IdentityGate, TwoQubitDiagonalGate

from qbraid.transpiler.conversions.qasm2.cirq_custom import RZZGate, U2Gate, U3Gate, rzz


def test_u2_gate():
    """Test U2Gate"""
    gate = U2Gate(0, 0)
    assert str(gate) == "U2(0,0)"
    assert gate._circuit_diagram_info_(None) == "U2(0.0, 0.0)"


def test_u3_gate():
    """Test U3Gate"""
    gate = U3Gate(0, 0, 0)
    assert str(gate) == "U3(0,0,0)"
    assert gate._circuit_diagram_info_(None) == "U3(0.0, 0.0, 0.0)"


class FakeArgs:
    """Fake Args class for testing"""

    def __init__(self, precision=None):
        self.precision = precision


def test_rzz_gate():
    """Test RZZGate"""
    rads = 0.2
    gate = RZZGate(rads)

    itheta2 = 1j * rads * np.pi / 2
    expected_unitary = np.array(
        [
            [np.exp(-itheta2), 0, 0, 0],
            [0, np.exp(itheta2), 0, 0],
            [0, 0, np.exp(itheta2), 0],
            [0, 0, 0, np.exp(-itheta2)],
        ],
    )
    assert gate._num_qubits_() == 2
    assert gate._unitary_().all() == expected_unitary.all()
    assert gate._circuit_diagram_info_(args=FakeArgs(1)).wire_symbols[0] == "RZZ(0.2)"


def test_rzz_gate_method():
    """Test rzz method"""
    id_gate = rzz(0)
    assert isinstance(id_gate, IdentityGate)

    diagonal = rzz(2)
    assert isinstance(diagonal, TwoQubitDiagonalGate)

    random_rzz = rzz(0.5)
    assert isinstance(random_rzz, RZZGate)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing unified interface for interacting with
various quantum provider APIs.

.. currentmodule:: qbraid.runtime

.. _runtime_data_type:

Data Types
------------

.. autosummary::
   :toctree: ../stubs/

    DeviceStatus
    JobStatus
    NoiseModel
    NoiseModelSet
    ValidationLevel

Functions
------------

.. autosummary::
   :toctree: ../stubs/

    load_job
    get_providers
    load_provider
    display_jobs_from_data

Classes
--------

.. autosummary::
   :toctree: ../stubs/

    RuntimeOptions
    TargetProfile
    QuantumDevice
    QuantumJob
    QuantumProvider
    Result
    ResultData
    GateModelResultData
    AhsResultData
    AhsShotResult
    AnnealingResultData

Exceptions
------------

.. autosummary::
   :toctree: ../stubs/

    JobStateError
    ProgramValidationError
    QbraidRuntimeError
    ResourceNotFoundError
    DeviceProgramTypeMismatchError
    JobLoaderError
    ProviderLoaderError

"""
import importlib
from typing import TYPE_CHECKING

from ._display import display_jobs_from_data
from .device import QuantumDevice
from .enums import DeviceStatus, JobStatus, ValidationLevel
from .exceptions import (
    DeviceProgramTypeMismatchError,
    JobStateError,
    ProgramValidationError,
    QbraidRuntimeError,
    ResourceNotFoundError,
)
from .job import QuantumJob
from .loader import JobLoaderError, ProviderLoaderError, get_providers, load_job, load_provider
from .noise import NoiseModel, NoiseModelSet
from .options import RuntimeOptions
from .profile import TargetProfile
from .provider import QuantumProvider
from .result import Result
from .result_data import (
    AhsResultData,
    AhsShotResult,
    AnnealingResultData,
    GateModelResultData,
    ResultData,
)

PROVIDERS = get_providers()

__all__ = [
    "QuantumDevice",
    "DeviceStatus",
    "JobStatus",
    "display_jobs_from_data",
    "load_job",
    "get_providers",
    "load_provider",
    "JobStateError",
    "ProgramValidationError",
    "QbraidRuntimeError",
    "ResourceNotFoundError",
    "DeviceProgramTypeMismatchError",
    "JobLoaderError",
    "ProviderLoaderError",
    "TargetProfile",
    "QuantumJob",
    "QuantumProvider",
    "RuntimeOptions",
    "NoiseModel",
    "NoiseModelSet",
    "Result",
    "ResultData",
    "GateModelResultData",
    "AhsResultData",
    "AhsShotResult",
    "AnnealingResultData",
    "ValidationLevel",
    "PROVIDERS",
]

_lazy = {
    "aws": [
        "BraketProvider",
        "BraketDevice",
        "BraketQuantumTask",
    ],
    "azure": ["AzureQuantumProvider", "AzureQuantumDevice", "AzureQuantumJob"],
    "ionq": [
        "IonQSession",
        "IonQProvider",
        "IonQDevice",
        "IonQJob",
    ],
    "oqc": [
        "OQCProvider",
        "OQCDevice",
        "OQCJob",
    ],
    "ibm": [
        "QiskitRuntimeProvider",
        "QiskitBackend",
        "QiskitJob",
    ],
    "native": [
        "Session",
        "QbraidSession",
        "QbraidClient",
        "QbraidProvider",
        "QbraidDevice",
        "QbraidJob",
        "QirRunner",
    ],
    "schemas": [],
}

if TYPE_CHECKING:
    from .aws import BraketDevice as BraketDevice
    from .aws import BraketProvider as BraketProvider
    from .aws import BraketQuantumTask as BraketQuantumTask
    from .azure import AzureQuantumDevice as AzureQuantumDevice
    from .azure import AzureQuantumJob as AzureQuantumJob
    from .azure import AzureQuantumProvider as AzureQuantumProvider
    from .ibm import QiskitBackend as QiskitBackend
    from .ibm import QiskitJob as QiskitJob
    from .ibm import QiskitRuntimeProvider as QiskitRuntimeProvider
    from .ionq import IonQDevice as IonQDevice
    from .ionq import IonQJob as IonQJob
    from .ionq import IonQProvider as IonQProvider
    from .ionq import IonQSession as IonQSession
    from .native import QbraidClient as QbraidClient
    from .native import QbraidDevice as QbraidDevice
    from .native import QbraidJob as QbraidJob
    from .native import QbraidProvider as QbraidProvider
    from .native import QbraidSession as QbraidSession
    from .native import QirRunner as QirRunner
    from .native import Session as Session
    from .oqc import OQCDevice as OQCDevice
    from .oqc import OQCJob as OQCJob
    from .oqc import OQCProvider as OQCProvider


def __getattr__(name):
    for mod_name, objects in _lazy.items():
        if name == mod_name:
            module = importlib.import_module(f".{mod_name}", __name__)
            globals()[mod_name] = module
            return module

        if name in objects:
            module = importlib.import_module(f".{mod_name}", __name__)
            obj = getattr(module, name)
            globals()[name] = obj
            return obj

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(
        __all__ + list(_lazy.keys()) + [item for sublist in _lazy.values() for item in sublist]
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests native runtime job schemas.

"""
from collections import Counter
from datetime import datetime
from decimal import Decimal

import pytest
from pydantic import BaseModel, ValidationError

from qbraid.programs import ExperimentType
from qbraid.runtime.enums import JobStatus
from qbraid.runtime.schemas.base import USD, Credits, QbraidSchemaBase, QbraidSchemaHeader
from qbraid.runtime.schemas.device import DeviceData, DevicePricing
from qbraid.runtime.schemas.experiment import (
    AhsExperimentMetadata,
    ExperimentMetadata,
    GateModelExperimentMetadata,
    QuboSolveParams,
)
from qbraid.runtime.schemas.job import RuntimeJobModel, TimeStamps


@pytest.fixture
def mock_job_data(valid_qasm2_no_meas):
    """Return mock job data."""
    return {
        "qbraidJobId": "job_123",
        "qbraidDeviceId": "device_456",
        "status": "COMPLETED",
        "shots": 100,
        "experimentType": "gate_model",
        "openQasm": valid_qasm2_no_meas,
        "circuitNumQubits": 2,
        "circuitDepth": 2,
        "timeStamps": {
            "createdAt": "2024-09-14T01:06:34.000Z",
            "endedAt": "2024-09-14T01:06:35.000Z",
            "executionDuration": 60000,
        },
    }


def test_qbraid_schema_header():
    """Test QbraidSchemaHeader class."""
    header = QbraidSchemaHeader(name="TestSchema", version=1.0)
    assert header.name == "TestSchema"
    assert header.version == 1.0

    with pytest.raises(ValidationError):
        QbraidSchemaHeader(name="", version=1.0)  # Invalid name

    with pytest.raises(ValidationError):
        QbraidSchemaHeader(name="TestSchema", version=0)  # Invalid version


def test_qbraid_schema_base_raises_for_no_version():
    """Test QbraidSchemaBase raises ValueError when accessing header property."""
    with pytest.raises(ValueError) as excinfo:
        _ = QbraidSchemaBase()
    assert "QbraidSchemaBase must define a valid semantic version for 'VERSION'." in str(
        excinfo.value
    )


def test_qbraid_schema_base():
    """Test QbraidSchemaBase class."""

    class TestSchema(QbraidSchemaBase):
        """Test schema class."""

        VERSION = 6.1

    model = TestSchema()
    assert model.header.name == "tests.runtime.test_schemas"
    assert model.header.version == 6.1


def test_gate_model_experiment_metadata(valid_qasm2_no_meas):
    """Test GateModelExperimentMetadata class."""
    metadata = GateModelExperimentMetadata(
        openQasm=valid_qasm2_no_meas,
        measurementCounts=Counter({"00": 10, "01": 15}),
    )
    assert metadata.qasm == valid_qasm2_no_meas
    assert metadata.num_qubits == 2
    assert metadata.gate_depth == 3
    assert metadata.measurement_counts == Counter({"00": 10, "01": 15})

    with pytest.raises(ValidationError):
        GateModelExperimentMetadata(openQasm="invalid qasm string")


def test_validate_counts():
    """Test validate_counts method."""
    valid_counts = Counter({"00": 10, "01": 20})
    validated_counts = GateModelExperimentMetadata.validate_counts(valid_counts)
    assert validated_counts == valid_counts

    invalid_counts = {"00": 10, "01": 20}
    validated_counts = GateModelExperimentMetadata.validate_counts(invalid_counts)
    assert isinstance(validated_counts, Counter)


def test_timestamps():
    """Test TimeStamps class."""
    created_at_obj = datetime(2024, 9, 14, 1, 6, 34, 0)
    ended_at_obj = datetime(2024, 9, 14, 1, 6, 35, 0)

    created_at_str = "2024-09-14T01:06:34.000Z"
    ended_at_str = "2024-09-14T01:06:35.000Z"

    duration = 60000

    time_stamps = TimeStamps(
        createdAt=created_at_str, endedAt=ended_at_str, executionDuration=duration
    )
    assert time_stamps.createdAt == created_at_obj
    assert time_stamps.endedAt == ended_at_obj
    assert time_stamps.executionDuration == duration

    with pytest.raises(ValidationError):
        TimeStamps(
            createdAt="invalid date", endedAt=ended_at_str, executionDuration=duration
        )  # Invalid date format

    with pytest.raises(ValidationError):
        TimeStamps(
            createdAt=created_at_str, endedAt=ended_at_str, executionDuration=-1
        )  # Invalid executionDuration


def test_execution_duration_auto_calculation():
    """Test execution duration auto-calculation."""
    created = datetime(2025, 5, 22, 15, 8, 57)
    ended = datetime(2025, 5, 22, 15, 8, 58)
    ts = TimeStamps(createdAt=created, endedAt=ended)
    assert ts.executionDuration == 1000  # 1 second in milliseconds


def test_execution_duration_none_if_endedAt_missing():
    """Test execution duration is None if endedAt is missing."""
    created = datetime(2025, 5, 22, 15, 8, 57)
    ts = TimeStamps(createdAt=created)
    assert ts.executionDuration is None  # Not enough data to compute


def test_execution_duration_zero_if_same_time():
    """Test execution duration is zero if createdAt and endedAt are the same."""
    t = datetime(2025, 5, 22, 15, 8, 57)
    ts = TimeStamps(createdAt=t, endedAt=t)
    assert ts.executionDuration == 0


def test_runtime_job_model(mock_job_data):
    """Test RuntimeJobModel class."""

    job = RuntimeJobModel.from_dict(mock_job_data)
    assert job.job_id == "job_123"
    assert job.device_id == "device_456"
    assert job.status == JobStatus.COMPLETED
    assert job.shots == 100
    assert isinstance(job.metadata, GateModelExperimentMetadata)
    assert job.time_stamps.executionDuration == 60000  # pylint: disable=no-member

    # Test job with missing metadata
    job_data = mock_job_data.copy()
    job_data.pop("metadata", None)
    job = RuntimeJobModel.from_dict(job_data)
    assert isinstance(job.metadata, GateModelExperimentMetadata)

    with pytest.raises(ValueError):
        RuntimeJobModel.from_dict(
            {"qbraidJobId": "job_123", "qbraidDeviceId": "device_456", "status": "invalid_status"}
        )


def test_populate_metadata(valid_qasm2_no_meas):
    """Test _populate_metadata method."""
    job_data_gm = {"openQasm": valid_qasm2_no_meas, "circuitNumQubits": 5, "circuitDepth": 10}
    populated_data = RuntimeJobModel._populate_metadata(job_data_gm, ExperimentType.GATE_MODEL)
    assert isinstance(populated_data["metadata"], GateModelExperimentMetadata)
    assert populated_data["metadata"].qasm == valid_qasm2_no_meas

    job_data_ahs = {"numAtoms": 10}
    populated_data = RuntimeJobModel._populate_metadata(job_data_ahs, ExperimentType.AHS)
    assert isinstance(populated_data["metadata"], AhsExperimentMetadata)


def test_runtime_job_model_metadata_population():
    """Test metadata population in RuntimeJobModel."""
    job_data = {
        "qbraidJobId": "job_123",
        "qbraidDeviceId": "device_456",
        "status": "CANCELLED",
        "shots": 100,
        "experimentType": "other",
        "someAdditionalField": "extra data",
        "timeStamps": {
            "createdAt": "2024-09-14T01:06:34.000Z",
            "endedAt": "2024-09-14T01:06:35.000Z",
            "executionDuration": 60000,
        },
    }
    job = RuntimeJobModel.from_dict(job_data)
    assert isinstance(job.metadata, ExperimentMetadata)
    assert job.metadata.someAdditionalField == "extra data"


def test_time_stamps_created_at_fallback():
    """Test fallback to createdAt for TimeStamps."""
    job_data = {
        "qbraidJobId": "job_123",
        "qbraidDeviceId": "device_456",
        "status": "RUNNING",
        "shots": 100,
        "experimentType": "ahs",
        "createdAt": "2024-09-14T01:06:34.000Z",
    }
    job = RuntimeJobModel.from_dict(job_data)

    # pylint: disable=no-member
    assert job.time_stamps.createdAt == datetime(2024, 9, 14, 1, 6, 34, 0)
    assert job.time_stamps.endedAt is None
    assert job.time_stamps.executionDuration is None
    # pylint: enable=no-member


def test_invalid_experiment_type_raises_error(mock_job_data):
    """Test invalid experiment type raises error."""
    bad_type = "unsupported type"
    bad_job_data = mock_job_data.copy()
    bad_job_data["experimentType"] = bad_type
    with pytest.raises(ValueError) as excinfo:
        RuntimeJobModel.from_dict(bad_job_data)
    assert str(excinfo.value).startswith(f"'{bad_type}' is not a valid ExperimentType")

    model = RuntimeJobModel.from_dict(mock_job_data)

    with pytest.raises(ValidationError) as excinfo:
        _ = RuntimeJobModel(
            qbraidJobId=model.job_id,
            qbraidDeviceId=model.device_id,
            status=model.status.value,
            statusText=model.status_text,
            shots=model.shots,
            experimentType=bad_type,
            metadata=model.metadata,
            timeStamps=model.time_stamps,
            tags=model.tags,
            cost=model.cost,
        )
    assert f"Invalid experiment_type: '{bad_type}'" in str(excinfo.value)


def test_invalid_status_raises_error(mock_job_data):
    """Test invalid job status raises error."""
    bad_type = "unsupported type"
    bad_job_data = mock_job_data.copy()
    bad_job_data["status"] = bad_type
    with pytest.raises(ValueError) as excinfo:
        RuntimeJobModel.from_dict(bad_job_data)
    assert str(excinfo.value).startswith(f"'{bad_type}' is not a valid JobStatus")

    with pytest.raises(ValidationError) as excinfo:
        RuntimeJobModel(**bad_job_data)
    assert f"Invalid status: '{bad_type}'" in str(excinfo.value)


def test_parse_datetimes_for_none_type(mock_job_data):
    """Test that TimeStamps model can handle None values."""
    time_stamps = mock_job_data["timeStamps"]
    time_stamps["endedAt"] = None
    time_stamps_model = TimeStamps(**time_stamps)
    assert time_stamps_model.endedAt is None


def test_device_data_qir(device_data_qir):
    """Test DeviceData class for qBraid QIR simulator."""
    device = DeviceData(**device_data_qir)

    assert device.provider == "qBraid"
    assert device.name == "QIR sparse simulator"
    assert device.paradigm == "gate-based"
    assert device.device_type == "SIMULATOR"
    assert device.num_qubits == 64
    assert device.pricing.perTask == Decimal("0.005")
    assert device.pricing.perShot == Decimal("0")
    assert device.pricing.perMinute == Decimal("0.075")
    assert device.status == "ONLINE"
    assert device.is_available is True


def test_device_data_quera(device_data_quera):
    """Test DeviceData class for QuEra QASM simulator."""
    device = DeviceData(**device_data_quera)

    assert device.provider == "QuEra"
    assert device.name == "Noisey QASM simulator"
    assert device.paradigm == "gate-based"
    assert device.device_type == "SIMULATOR"
    assert device.num_qubits == 30
    assert device.pricing.perTask == Decimal("0")
    assert device.pricing.perShot == Decimal("0")
    assert device.pricing.perMinute == Decimal("0")
    assert device.status == "ONLINE"
    assert device.is_available is True


def test_device_data_aquila(device_data_aquila):
    """Test DeviceData class for QuEra Aquila QPU."""
    device = DeviceData(**device_data_aquila)

    assert device.provider == "QuEra"
    assert device.name == "Aquila"
    assert device.paradigm == "AHS"
    assert device.device_type == "QPU"
    assert device.num_qubits == 256
    assert device.pricing.perTask == Decimal("0.3")
    assert device.pricing.perShot == Decimal("0.01")
    assert device.pricing.perMinute == Decimal("0")
    assert device.status == "OFFLINE"
    assert device.is_available is False


def test_usd_json_schema():
    """Test JSON schema for USD type."""

    class TestModel(BaseModel):
        """Mock model with USD field."""

        amount: USD

    json_schema = TestModel.model_json_schema()
    amount_schema = json_schema["properties"]["amount"]

    assert amount_schema["title"] == "USD"
    assert amount_schema["description"] == "A monetary amount representing U.S. Dollars."
    assert amount_schema["examples"] == [10, 0.05, 1.5]
    assert amount_schema["type"] == "number"


def test_credits_json_schema():
    """Test JSON schema for Credits type."""

    class TestModel(BaseModel):
        """Mock model with Credits field."""

        amount: Credits

    json_schema = TestModel.model_json_schema()
    amount_schema = json_schema["properties"]["amount"]

    assert amount_schema["title"] == "Credits"
    assert amount_schema["description"] == "A monetary amount where 1 Credit = $0.01 USD."
    assert amount_schema["examples"] == [10, 0.05, 1.5]
    assert amount_schema["type"] == "number"


def test_credits_pydantic_parsing():
    """Test parsing of Credits type in Pydantic models."""

    class TestModel(BaseModel):
        """Mock model with Credits field."""

        amount: Credits

    model_int = TestModel(amount=100)
    assert isinstance(model_int.amount, Credits)
    assert model_int.amount == Credits(100)

    model_float = TestModel(amount=100.5)
    assert isinstance(model_float.amount, Credits)
    assert model_float.amount == Credits(100.5)

    model_decimal = TestModel(amount=Decimal("100.75"))
    assert isinstance(model_decimal.amount, Credits)
    assert model_decimal.amount == Credits(Decimal("100.75"))

    model_str = TestModel(amount="200")
    assert isinstance(model_str.amount, Credits)
    assert model_str.amount == Credits(200)

    with pytest.raises(ValidationError) as excinfo:
        TestModel(amount="invalid")

    assert "Input should be a number (int, float, or Decimal)" in str(excinfo.value)


def test_qubo_solve_params_model():
    """Test QuboSolveParams with valid parameters."""
    params = QuboSolveParams(
        offset=0.5,
        num_reads=10,
        num_results=5,
        num_sweeps=1000,
        beta_range=(5.0, 100.0, 200),
        beta_list=[1.2, 5.5, 10.0],
        dense=True,
        vector_mode="speed",
        timeout=3600,
        ve_num=2,
        onehot=[["x[0]", "x[1]"]],
        fixed={"x[0]": 1, "x[1]": 0},
        andzero=[["x[0]", "x[1]"]],
        orone=[["x[1]", "x[2]"]],
        supplement=[["y[0]", "y[1]"]],
        maxone=[[1, ["x[0]", "x[1]"]]],
        minmaxone=[[1, 2, ["x[1]", "x[2]"]]],
        init_spin={"x[0]": 1, "x[1]": 0},
        spin_list=["x[1]", "x[2]"],
    )
    assert params.num_reads == 10
    assert params.beta_range == (5.0, 100.0, 200)
    assert params.vector_mode == "speed"


def test_qubo_solve_params_invalid_offset():
    """Test QuboSolveParams with an invalid offset value outside of the allowed range."""
    with pytest.raises(ValidationError, match="offset must be between"):
        QuboSolveParams(offset=3.5e40)


def test_qubo_solve_params_invalid_num_reads():
    """Test QuboSolveParams with invalid num_reads outside the range of 1 to 20."""
    with pytest.raises(ValidationError, match="num_reads must be between 1 and 20"):
        QuboSolveParams(offset=0.5, num_reads=25)


def test_qubo_solve_params_invalid_num_sweeps():
    """Test QuboSolveParams with invalid num_sweeps outside the range of 1 to 100000."""
    with pytest.raises(ValidationError, match="num_sweeps must be between 1 and 100000"):
        QuboSolveParams(offset=0.5, num_sweeps=200000)


def test_qubo_solve_params_invalid_beta_range():
    """Test QuboSolveParams with invalid beta_range values."""
    with pytest.raises(ValidationError, match="start value must be between"):
        QuboSolveParams(offset=0.5, beta_range=(1e-40, 100.0, 200))

    with pytest.raises(ValidationError, match="end value must be between"):
        QuboSolveParams(offset=0.5, beta_range=(10.0, 4e40, 200))

    with pytest.raises(
        ValidationError, match="start value must be less than or equal to end value"
    ):
        QuboSolveParams(offset=0.5, beta_range=(200.0, 10.0, 200))

    with pytest.raises(ValidationError, match="steps must be between 1 and 100000"):
        QuboSolveParams(offset=0.5, beta_range=(10.0, 100.0, -2))


def test_qubo_solve_params_invalid_beta_list():
    """Test QuboSolveParams with invalid beta_list values outside the allowed range."""
    with pytest.raises(ValidationError, match="All beta values must be between"):
        QuboSolveParams(offset=0.5, beta_list=[1e-40, 10.0, 5e40])


def test_qubo_solve_params_invalid_timeout():
    """Test QuboSolveParams with an invalid timeout outside the range of 1 to 7200."""
    with pytest.raises(ValidationError, match="timeout must be between 1 and 7200 seconds"):
        QuboSolveParams(offset=0.5, timeout=8000)


def test_qubo_solve_params_invalid_vector_mode():
    """Test QuboSolveParams with an invalid vector_mode value."""
    with pytest.raises(ValidationError, match="vector_mode must be 'speed' or 'accuracy'"):
        QuboSolveParams(offset=0.5, vector_mode="fast")


def test_qubo_solve_params_invalid_ve_num():
    """Test QuboSolveParams with invalid ve_num below 1."""
    with pytest.raises(ValidationError, match="ve_num must be greater than or equal to 1"):
        QuboSolveParams(offset=0.5, ve_num=0)


def test_qubo_solve_params_beta_list_none():
    """Test QuboSolveParams beta_list set to None."""
    params = QuboSolveParams(offset=0.5, beta_list=None)
    assert params.beta_list is None


def test_device_pricing():
    """Test DevicePricing class."""
    pricing = DevicePricing(
        perTask=USD(0.5),
        perShot=USD(0.01),
        perMinute=USD(0.075),
    )
    assert pricing.perTask == USD(0.5)
    assert pricing.perShot == USD(0.01)
    assert pricing.perMinute == USD(0.075)

    pricing_credits = pricing.serialize_credits(pricing.perTask)
    assert isinstance(pricing_credits, Credits)
    assert pricing_credits == Credits(50)


@pytest.mark.parametrize(
    "sites, expected",
    [
        ([(1.0, 2.0), (3.0, 4.0)], [(1.0, 2.0), (3.0, 4.0)]),
        ([(1, 2), (3, 4)], [(1.0, 2.0), (3.0, 4.0)]),
        (None, None),
        ([], []),
    ],
)
def test_ahs_expt_validate_sites_valid(sites, expected):
    """Test AhsExperimentMetadata sites validation."""
    model = AhsExperimentMetadata(sites=sites)
    assert model.sites == expected


@pytest.mark.parametrize(
    "sites",
    [
        [(1, 2, 3)],
        [(1,)],
        [1, 2],
        ["a", "b"],
        [("a", "b")],
        "not a list",
        [("1.0", "2.0"), ("3.0", "not_a_number")],
        [(1, 2, 3), (4, 5)],
        [(1, 2), "not_a_tuple"],
    ],
)
def test_ahs_expt_validate_sites_invalid(sites):
    """Test raising ValidationError for invalid sites."""
    with pytest.raises(ValidationError):
        AhsExperimentMetadata(sites=sites)


def test_ahs_expt_sites_lists_to_tuples():
    """Test AhsExperimentMetadata sites lists converted to tuples."""
    model = AhsExperimentMetadata(sites=[[1, 2], [3, 4]])
    assert model.sites == [(1.0, 2.0), (3.0, 4.0)]


def test_ahs_expt_measurement_counts_dict_to_counter():
    """Test AhsExperimentMetadata measurement_counts dict converted to Counter."""
    model = AhsExperimentMetadata(measurementCounts={"ggrggrgg": 1, "rgggrggg": 1})
    assert model.measurement_counts == Counter({"ggrggrgg": 1, "rgggrggg": 1})


@pytest.mark.parametrize(
    "filling, expected",
    [
        ([0, 1, 0, 1], [0, 1, 0, 1]),
        ([True, False, True], [1, 0, 1]),
        (None, None),
        ([], []),
    ],
)
def test_ahs_expt_validate_filling_valid(filling, expected):
    """Test AhsExperimentMetadata filling validation."""
    model = AhsExperimentMetadata(filling=filling)
    assert model.filling == expected


@pytest.mark.parametrize(
    "filling",
    [[0, 1, 2], [0, 1, -1], [0.5, 1.5], "not a list"],
)
def test_ahs_expt_validate_filling_invalid(filling):
    """Test raising ValidationError for invalid filling."""
    with pytest.raises(ValidationError):
        AhsExperimentMetadata(filling=filling)


def test_ahs_expt_validate_infer_num_atoms():
    """Test inferring num_atoms from sites and filling."""
    model = AhsExperimentMetadata(sites=[(0, 0), (1, 1), (2, 2)], filling=[0, 1, 0])
    assert model.num_atoms == 3


def test_ahs_expt_validate_sites_filling_mismatch():
    """Test raising ValidationError for sites and filling length mismatch."""
    with pytest.raises(
        ValidationError,
        match="The lengths of 'sites', 'filling', and value of 'num_atoms' must be consistent.",
    ):
        AhsExperimentMetadata(sites=[(0, 0), (1, 1)], filling=[0, 1, 0])


def test_ahs_expt_validate_num_atoms_mismatch():
    """Test raising ValidationError for num_atoms mismatch."""
    with pytest.raises(
        ValidationError,
        match="The lengths of 'sites', 'filling', and value of 'num_atoms' must be consistent.",
    ):
        AhsExperimentMetadata(sites=[(0, 0), (1, 1)], filling=[0, 1], numAtoms=3)


@pytest.mark.parametrize(
    "sites, filling",
    [
        ([(0, 0), (1, 1)], None),
        (None, [0, 1]),
    ],
)
def test_ahs_expt_validate_partial_data(sites, filling):
    """Test AhsExperimentMetadata with partial data."""
    model = AhsExperimentMetadata(sites=sites, filling=filling)
    assert model.num_atoms == 2


def test_ahs_expt_validate_empty_lists():
    """Test AhsExperimentMetadata with empty lists."""
    model = AhsExperimentMetadata(sites=[], filling=[])
    assert model.num_atoms is None



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
OpenQASM 2 conversions

.. currentmodule:: qbraid.transpiler.conversions.qasm2

Functions
----------

.. autosummary::
    :toctree: ../stubs/

    qasm2_to_cirq
    qasm2_to_pytket
    qasm2_to_qiskit
    qasm2_to_qasm3
    qasm2_to_ionq
    qasm2_to_qibo
    qibo_to_qasm2
    qasm2_to_pyqpanda3
    pyqpanda3_to_qasm2

"""

from .qasm2_extras import pyqpanda3_to_qasm2, qasm2_to_pyqpanda3, qasm2_to_qibo, qibo_to_qasm2
from .qasm2_to_cirq import qasm2_to_cirq
from .qasm2_to_ionq import qasm2_to_ionq
from .qasm2_to_pytket import qasm2_to_pytket
from .qasm2_to_qasm3 import qasm2_to_qasm3
from .qasm2_to_qiskit import qasm2_to_qiskit

__all__ = [
    "qasm2_to_cirq",
    "qasm2_to_pytket",
    "qasm2_to_qasm3",
    "qasm2_to_qiskit",
    "qasm2_to_ionq",
    "qasm2_to_qibo",
    "qibo_to_qasm2",
    "qasm2_to_pyqpanda3",
    "pyqpanda3_to_qasm2",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing functions to convert between OpenQASM 2 and IonQ JSON format.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid.transpiler.annotations import weight
from qbraid.transpiler.conversions.openqasm3.openqasm3_to_ionq import openqasm3_to_ionq

if TYPE_CHECKING:
    from qbraid.programs.typer import IonQDictType, Qasm2StringType


@weight(1)
def qasm2_to_ionq(qasm: Qasm2StringType) -> IonQDictType:
    """Returns an IonQ JSON format representation the input OpenQASM 2 string.

    Args:
        qasm (str): OpenQASM 2 string to convert to IonQDict type.

    Returns:
        dict: IonQ JSON format equivalent to input OpenQASM 2 string.
    """
    return openqasm3_to_ionq(qasm)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM to QASM 3 conversion function

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import openqasm3

from qbraid.transpiler.annotations import weight

if TYPE_CHECKING:
    from qbraid.programs.typer import Qasm3StringType


@weight(1)
def openqasm3_to_qasm3(program: openqasm3.ast.Program) -> Qasm3StringType:
    """Dumps openqasm3.ast.Program to an OpenQASM 3.0 string

    Args:
        program (openqasm3.ast.Program): OpenQASM 3.0 AST program

    Returns:
        str: OpenQASM 3.0 string
    """
    statements = program.statements
    program = openqasm3.ast.Program(statements=statements, version="3.0")
    return openqasm3.dumps(program)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Benchmarking tests for pyQuil conversions

"""
import importlib.util
import string

import numpy as np
import pytest

from qbraid.interface import circuits_allclose
from qbraid.transpiler import ConversionGraph, transpile

try:
    import pyquil

    pyquil_not_installed = False
except ImportError:
    pyquil_not_installed = True

pytestmark = pytest.mark.skipif(pyquil_not_installed, reason="pyquil not installed")


def generate_params(varnames, seed=0):
    """Generate random parameters to help construct PyQuil test gates"""
    np.random.seed(seed)
    params = {}
    rot_args = ["angle", "phi", "lam", "gamma"]
    for ra in rot_args:
        if ra in varnames:
            params[ra] = np.random.rand() * 2 * np.pi

    if "qubit" in varnames:
        params["qubit"] = 0

    if "control" in varnames:
        params["control"] = 0
        params["target"] = 1

    if "q1" in varnames:
        params["q1"] = 0
        params["q2"] = 1

    return params


@pytest.fixture
def pyquil_gates():
    """Construct a dictionary of all pyQuil gates and assign random parameters as applicable"""
    pyquil_gate_dict = {
        attr: None for attr in dir(pyquil.gates) if attr[0] in string.ascii_uppercase
    }

    for gate in pyquil_gate_dict:
        try:
            params = generate_params(getattr(pyquil.gates, gate).__code__.co_varnames)
            pyquil_gate_dict[gate] = getattr(pyquil.gates, gate)(**params)
        except Exception:  # pylint: disable=broad-exception-caught
            continue

    return {k: v for k, v in pyquil_gate_dict.items() if v is not None}


@pytest.fixture
def conversion_graph():
    """Construct a conversion graph with native support for all target program types"""
    return ConversionGraph(require_native=True)


def is_package_installed(package_name: str) -> bool:
    """Check if a package is installed."""
    return importlib.util.find_spec(package_name) is not None


ALL_TARGETS = [("braket", 0.77), ("cirq", 0.77), ("pytket", 0.77), ("qiskit", 0.77)]
AVAILABLE_TARGETS = [(name, version) for name, version in ALL_TARGETS if is_package_installed(name)]


def convert_from_pyquil_to_x(target, gate_name, gates, graph):
    """Construct a pyQuil programs with the given gate, transpile it to
    target program type, and check equivalence.
    """
    gate = gates[gate_name]
    source_circuit = pyquil.Program()
    source_circuit += gate
    target_circuit = transpile(source_circuit, target, conversion_graph=graph)
    assert circuits_allclose(source_circuit, target_circuit, strict_gphase=False)


@pytest.mark.parametrize(("target", "baseline"), AVAILABLE_TARGETS)
def test_pyquil_coverage(target, baseline, pyquil_gates, conversion_graph):
    """Test converting pyQuil programs to supported target program type over
    all pyQuil gates and check against baseline expecte accuracy.
    """
    ACCURACY_BASELINE = baseline
    ALLOWANCE = 0.01
    failures = {}
    for gate_name in pyquil_gates:
        try:
            convert_from_pyquil_to_x(target, gate_name, pyquil_gates, conversion_graph)
        except Exception as e:  # pylint: disable=broad-exception-caught
            failures[f"{target}-{gate_name}"] = e

    total_tests = len(pyquil_gates)
    nb_fails = len(failures)
    nb_passes = total_tests - nb_fails
    accuracy = float(nb_passes) / float(total_tests)

    assert accuracy >= ACCURACY_BASELINE - ALLOWANCE, (
        f"The coverage threshold was not met. {nb_fails}/{total_tests} tests failed "
        f"({nb_fails / (total_tests):.2%}) and {nb_passes}/{total_tests} passed "
        f"(expected >= {ACCURACY_BASELINE}).\nFailures: {failures.keys()}\n\n"
    )



# Copyright (C) 2024 qBraid
# Copyright (C) Unitary Fund
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# This file includes code adapted from Mitiq (https://github.com/unitaryfund/mitiq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module for converting Braket circuits to Cirq circuits

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from braket.circuits import Circuit as BKCircuit
from braket.circuits import Instruction as BKInstruction
from braket.circuits import gates as braket_gates
from braket.circuits import noises as braket_noise_gate

try:
    import cirq
except ImportError:  # pragma: no cover
    cirq = None

try:
    import cirq_ionq.ionq_native_gates as cirq_ionq_ops
except ImportError:  # pragma: no cover
    cirq_ionq_ops = None

from qbraid.transpiler.annotations import weight
from qbraid.transpiler.exceptions import ProgramConversionError

if TYPE_CHECKING:
    import cirq.circuits as cirq_circuits
    import cirq.devices as cirq_devices
    import cirq.ops as cirq_ops


def _give_cirq_gate_name(gate: cirq_ops.Gate, name: str, n_qubits: int) -> cirq_ops.Gate:
    def _circuit_diagram_info_(args):  # pylint: disable=unused-argument
        return name, *(name,) * (n_qubits - 1)

    gate._circuit_diagram_info_ = _circuit_diagram_info_


def matrix_to_cirq_gate(matrix: np.ndarray) -> cirq_ops.MatrixGate:
    """Return cirq matrix gate given unitary"""
    n_qubits = int(np.log2(len(matrix)))
    unitary_gate = cirq.MatrixGate(matrix)
    _give_cirq_gate_name(unitary_gate, "U", n_qubits)
    return unitary_gate


def braket_gate_to_matrix(gate: braket_gates.Unitary) -> np.ndarray:
    """Return the matrix representation of a Braket gate."""
    matrix = gate.to_matrix()
    unitary_gate = braket_gates.Unitary(matrix)
    nqubits = int(np.log2(len(matrix)))
    qubits = list(range(nqubits)) if nqubits > 1 else 0
    bk_circuit = BKCircuit([BKInstruction(unitary_gate, qubits)])
    return bk_circuit.to_unitary()


@weight(0.99)
def braket_to_cirq(circuit: BKCircuit) -> cirq_circuits.Circuit:
    """Returns a Cirq circuit equivalent to the input Braket circuit.

    Note: The returned Cirq circuit acts on cirq.LineQubit's with indices equal
    to the qubit indices of the Braket circuit.

    Args:
        circuit: Braket circuit to convert to a Cirq circuit.

    Returns:
        Cirq circuit equivalent to the input Braket circuit.
    """
    # pylint: disable-next=import-outside-toplevel
    from qbraid.programs.gate_model.cirq import CirqCircuit as QbraidCircuit

    bk_qubits = [int(q) for q in circuit.qubits]
    cirq_qubits = [cirq.LineQubit(x) for x in bk_qubits]
    qubit_mapping = {q: cirq_qubits[i] for i, q in enumerate(bk_qubits)}
    circuit = cirq.Circuit(
        _from_braket_instruction(instr, qubit_mapping) for instr in circuit.instructions
    )
    return QbraidCircuit.align_final_measurements(circuit)


def _from_braket_instruction(
    instr: BKInstruction, qubit_mapping: dict[int, cirq_devices.LineQubit]
) -> list[cirq_ops.Operation]:
    """Converts the braket instruction to an equivalent Cirq operation or list
    of Cirq operations.

    Args:
        instr: Braket instruction to convert.
        qubit_mapping: Braket qubit indicies mapped to indexed Cirq LineQubits

    Raises:
        ValueError: If the instruction cannot be converted to Cirq.
        ProgramConversionError: If error raise during conversion.
    """
    nqubits = len(instr.target)
    BK_qubits = [int(q) for q in instr.target]
    qubits = [qubit_mapping[x] for x in BK_qubits]

    if str(instr.operator) == "Measure":
        return [cirq.ops.MeasurementGate(num_qubits=nqubits).on(*qubits)]

    try:
        if nqubits == 1:
            return _from_one_qubit_braket_instruction(instr, qubits)

        if nqubits == 2:
            return _from_two_qubit_braket_instruction(instr, qubits)

        if nqubits == 3:
            if isinstance(instr.operator, braket_gates.CCNot):
                return [cirq.ops.TOFFOLI.on(*qubits)]
            if isinstance(instr.operator, braket_gates.CSwap):
                return [cirq.ops.FREDKIN.on(*qubits)]
            try:
                matrix = braket_gate_to_matrix(instr.operator)
                return [cirq.ops.MatrixGate(matrix).on(*qubits)]
            except (ValueError, TypeError) as err:
                raise ProgramConversionError(
                    f"Unable to convert the instruction {instr} to Cirq."
                ) from err

        # Unknown instructions.
        raise ProgramConversionError(
            f"Unable to convert to Cirq due to unrecognized \
            instruction: {instr}."
        )

    except Exception as err:
        raise ProgramConversionError(
            f"qBraid transpiler doesn't support operator {instr.operator}"
        ) from err


def _from_one_qubit_braket_instruction(
    instr: BKInstruction, qubits: list[cirq_devices.LineQubit]
) -> list[cirq_ops.Operation]:
    """Converts the one-qubit Braket instruction to Cirq operation(s).

    Args:
        instr: One-qubit Braket instruction to convert.
        qubits: Cirq LineQubit list indexed according to Braket instruction

    Raises:
        ValueError: If the instruction cannot be converted to Cirq.
    """
    gate = instr.operator

    # One-qubit non-parameterized gates.
    if isinstance(gate, braket_gates.I):
        return [cirq.ops.I.on(*qubits)]
    if isinstance(gate, braket_gates.X):
        return [cirq.ops.X.on(*qubits)]
    if isinstance(gate, braket_gates.Y):
        return [cirq.ops.Y.on(*qubits)]
    if isinstance(gate, braket_gates.Z):
        return [cirq.ops.Z.on(*qubits)]
    if isinstance(gate, braket_gates.H):
        return [cirq.ops.H.on(*qubits)]
    if isinstance(gate, braket_gates.S):
        return [cirq.ops.S.on(*qubits)]
    if isinstance(gate, braket_gates.Si):
        return [cirq.protocols.inverse(cirq.ops.S.on(*qubits))]
    if isinstance(gate, braket_gates.T):
        return [cirq.ops.T.on(*qubits)]
    if isinstance(gate, braket_gates.Ti):
        return [cirq.protocols.inverse(cirq.ops.T.on(*qubits))]
    if isinstance(gate, braket_gates.V):
        return [cirq.ops.X.on(*qubits) ** 0.5]
    if isinstance(gate, braket_gates.Vi):
        return [cirq.ops.X.on(*qubits) ** -0.5]

    # One-qubit parameterized gates.
    if isinstance(gate, braket_gates.Rx):
        return [cirq.ops.rx(gate.angle).on(*qubits)]
    if isinstance(gate, braket_gates.Ry):
        return [cirq.ops.ry(gate.angle).on(*qubits)]
    if isinstance(gate, braket_gates.Rz):
        return [cirq.ops.rz(gate.angle).on(*qubits)]
    if isinstance(gate, braket_gates.PhaseShift):
        return [cirq.ops.Z.on(*qubits) ** (gate.angle / np.pi)]

    # One-qubit parameterized IonQ gates
    if cirq_ionq_ops and isinstance(gate, (braket_gates.GPi, braket_gates.GPi2)):
        phi = gate.angle / (2 * np.pi)
        gate_class = (
            cirq_ionq_ops.GPIGate if isinstance(gate, braket_gates.GPi) else cirq_ionq_ops.GPI2Gate
        )
        return [gate_class(phi=phi).on(*qubits)]

    # One-qubit Noise gates.
    if isinstance(gate, braket_noise_gate.BitFlip):
        return [cirq.ops.BitFlipChannel(gate.probability).on(*qubits)]
    if isinstance(gate, braket_noise_gate.PhaseFlip):
        return [cirq.ops.PhaseFlipChannel(gate.probability).on(*qubits)]
    if isinstance(gate, braket_noise_gate.Depolarizing):
        return [cirq.ops.DepolarizingChannel(gate.probability).on(*qubits)]
    if isinstance(gate, braket_noise_gate.AmplitudeDamping):
        return [cirq.ops.AmplitudeDampingChannel(gate.gamma).on(*qubits)]
    if isinstance(gate, braket_noise_gate.GeneralizedAmplitudeDamping):
        return [
            cirq.ops.GeneralizedAmplitudeDampingChannel(gate.probability, gate.gamma).on(*qubits)
        ]
    if isinstance(gate, braket_noise_gate.PhaseDamping):
        return [cirq.ops.PhaseDampingChannel(gate.gamma).on(*qubits)]

    try:
        matrix = braket_gate_to_matrix(gate)
        return [cirq.ops.MatrixGate(matrix).on(*qubits)]
    except (ValueError, TypeError) as err:
        raise ValueError(f"Unable to convert the instruction {instr} to Cirq.") from err


def _from_two_qubit_braket_instruction(
    instr: BKInstruction, qubits: list[cirq_devices.LineQubit]
) -> list[cirq_ops.Operation]:
    """Converts the two-qubit braket instruction to Cirq.

    Args:
        instr: Two-qubit Braket instruction to convert.
        qubits: Cirq LineQubit list indexed according to Braket instruction

    Raises:
        ValueError: If the instruction cannot be converted to Cirq.
    """
    gate = instr.operator

    # Two-qubit non-parameterized gates.
    if isinstance(gate, braket_gates.CNot):
        return [cirq.ops.CNOT.on(*qubits)]

    if isinstance(gate, braket_gates.Swap):
        return [cirq.ops.SWAP.on(*qubits)]
    if isinstance(gate, braket_gates.ISwap):
        return [cirq.ops.ISWAP.on(*qubits)]
    if isinstance(gate, braket_gates.CZ):
        return [cirq.ops.CZ.on(*qubits)]
    if isinstance(gate, braket_gates.CY):
        return [
            cirq.protocols.inverse(cirq.ops.S.on(qubits[1])),
            cirq.ops.CNOT.on(*qubits),
            cirq.ops.S.on(qubits[1]),
        ]

    # Two-qubit parameterized gates.
    if isinstance(gate, braket_gates.CPhaseShift):
        return [cirq.ops.CZ.on(*qubits) ** (gate.angle / np.pi)]
    if isinstance(gate, braket_gates.CPhaseShift00):
        return [
            cirq.ops.XX(*qubits),
            cirq.ops.CZ.on(*qubits) ** (gate.angle / np.pi),
            cirq.ops.XX(*qubits),
        ]
    if isinstance(gate, braket_gates.CPhaseShift01):
        return [
            cirq.ops.X(qubits[0]),
            cirq.ops.CZ.on(*qubits) ** (gate.angle / np.pi),
            cirq.ops.X(qubits[0]),
        ]
    if isinstance(gate, braket_gates.CPhaseShift10):
        return [
            cirq.ops.X(qubits[1]),
            cirq.ops.CZ.on(*qubits) ** (gate.angle / np.pi),
            cirq.ops.X(qubits[1]),
        ]
    if isinstance(gate, braket_gates.PSwap):
        return [
            cirq.ops.SWAP.on(*qubits),
            cirq.ops.CNOT.on(*qubits),
            cirq.ops.Z.on(qubits[1]) ** (gate.angle / np.pi),
            cirq.ops.CNOT.on(*qubits),
        ]
    if isinstance(gate, braket_gates.XX):
        return [cirq.ops.XXPowGate(exponent=gate.angle / np.pi, global_shift=-0.5).on(*qubits)]
    if isinstance(gate, braket_gates.YY):
        return [cirq.ops.YYPowGate(exponent=gate.angle / np.pi, global_shift=-0.5).on(*qubits)]
    if isinstance(gate, braket_gates.ZZ):
        return [cirq.ops.ZZPowGate(exponent=gate.angle / np.pi, global_shift=-0.5).on(*qubits)]
    if isinstance(gate, braket_gates.XY):
        return [cirq.ops.ISwapPowGate(exponent=gate.angle / np.pi).on(*qubits)]

    # Two-qubit noise gates.
    if isinstance(gate, braket_noise_gate.Kraus):
        return [cirq.ops.KrausChannel(gate._matrices).on(*qubits)]
    if isinstance(gate, braket_noise_gate.TwoQubitDepolarizing):
        return [cirq.ops.DepolarizingChannel(gate.probability, n_qubits=2).on(*qubits)]

    # Two-qubit two-parameters IonQ gates.
    if cirq_ionq_ops and isinstance(gate, braket_gates.MS):
        return [
            cirq_ionq_ops.MSGate(
                phi0=gate.angle_1 / (2 * np.pi),
                phi1=gate.angle_2 / (2 * np.pi),
                theta=gate.angle_3 / (2 * np.pi),
            ).on(*qubits)
        ]

    try:
        matrix = braket_gate_to_matrix(gate)
        unitary_gate = matrix_to_cirq_gate(matrix)
        return [unitary_gate.on(*qubits)]
    except (ValueError, TypeError) as err:
        raise ValueError(f"Unable to convert the instruction {instr} to Cirq.") from err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for defining custom conversions

"""
from unittest.mock import Mock

import cirq
import numpy as np
import pytest

from qbraid.interface.random import random_circuit
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.transpiler.annotations import requires_extras, weight
from qbraid.transpiler.conversions.braket import braket_to_cirq
from qbraid.transpiler.edge import Conversion
from qbraid.transpiler.graph import ConversionGraph


@requires_extras("alice", "bob")
def dummy_func():
    """Dummy function for testing requires_extras decorator."""


def test_requires_extras_appends_dependency():
    """
    Test that the requires_extras decorator correctly appends a dependency
    to the function's attribute list.
    """
    assert getattr(dummy_func, "requires_extras") == ["alice", "bob"]


def test_raise_for_unsupported_program_input():
    """Test that an exception is raised for an unsupported program input."""
    conversion = Conversion("braket", "cirq", braket_to_cirq)
    circuit = Mock()
    with pytest.raises(ProgramTypeError):
        conversion.convert(circuit)


def test_raise_for_source_program_input_mismatch():
    """Test that an exception is raised for a mismatch between the source package and input."""
    source, target, other = "braket", "cirq", "qiskit"
    conversion = Conversion(source, target, braket_to_cirq)
    circuit = random_circuit(other)
    with pytest.raises(ValueError):
        conversion.convert(circuit)


def test_braket_cirq_custom_conversion():
    """Test that a Braket to Cirq conversion is successful."""
    source, target = "braket", "cirq"
    conversion = Conversion(source, target, braket_to_cirq)
    braket_circuit = random_circuit(source)
    cirq_circuit = conversion.convert(braket_circuit)
    assert isinstance(cirq_circuit, cirq.Circuit)


def test_conversion_repr():
    """Test that the conversion repr is correct."""
    source, target = "braket", "cirq"
    conversion = Conversion(source, target, braket_to_cirq)
    assert repr(conversion) == f"('{source}', '{target}')"


def test_valid_weights():
    """Test that valid weights are applied correctly."""

    @weight(0.5)
    def conversion_function():
        return "Converted"

    assert hasattr(conversion_function, "weight")
    assert conversion_function.weight == 0.5


def test_invalid_weights():
    """Test that weights outside the 0 to 1 range raise a ValueError."""

    with pytest.raises(ValueError) as excinfo:

        @weight(-0.1)
        def negative_weight_function():
            return "Should not work"

    assert "between 0 and 1" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:

        @weight(1.1)
        def too_high_weight_function():
            return "Should not work"

    assert "between 0 and 1" in str(excinfo.value)


def test_functionality():
    """Test that the decorated function works as expected."""

    @weight(1)
    def conversion_function():
        return "Functionality intact"

    assert conversion_function() == "Functionality intact"


def test_default_weight_application():
    """Test the application of a weight at the edge of the valid range."""

    @weight(1)
    def edge_case_function():
        return "Edge case test"

    assert edge_case_function.weight == 1
    assert edge_case_function() == "Edge case test"


@pytest.fixture
def mock_conversion_func():
    """Mock conversion function with a weight attribute."""

    def conversion_function():
        pass

    setattr(conversion_function, "weight", 0.5)

    return conversion_function


@pytest.mark.parametrize("invalid_weight", [-0.1, 1.1])
def test_invalid_weight(mock_conversion_func, invalid_weight):
    """Test initializing with a weight below zero."""
    with pytest.raises(ValueError):
        Conversion("source_pkg", "target_pkg", mock_conversion_func, weight=invalid_weight)


@pytest.mark.parametrize(
    "valid_weight,expected_value", [(0.8, np.log(1.25)), (None, np.log(2)), (0, float("inf"))]
)
def test_valid_weight(mock_conversion_func, valid_weight, expected_value):
    """Test the default weight from the conversion function if not specified."""
    conversion = Conversion("source_pkg", "target_pkg", mock_conversion_func, weight=valid_weight)
    assert conversion.weight == expected_value


def test_weight_without_specified_and_no_default():
    """Test default weight when no weight is specified and no default is in the function."""

    def simple_conversion_func():
        pass  # No weight attribute

    conversion = Conversion("source_pkg", "target_pkg", simple_conversion_func)
    assert conversion.weight == 0


class TestConversionEquality:
    """Tests for the Conversion class __eq__ method."""

    def test_equal_instances(self):
        """Test that two instances with identical initialization are considered equal."""
        conv1 = Conversion("source1", "target1", mock_conversion_func)
        conv2 = Conversion("source1", "target1", mock_conversion_func)
        assert conv1 == conv2, "Instances with identical initialization should be equal"

    def test_unequal_sources(self):
        """Test that instances with different source attributes are not equal."""
        conv1 = Conversion("source1", "target1", mock_conversion_func)
        conv2 = Conversion("source2", "target1", mock_conversion_func)
        assert conv1 != conv2, "Instances with different sources should not be equal"

    def test_unequal_targets(self):
        """Test that instances with different target attributes are not equal."""
        conv1 = Conversion("source1", "target1", mock_conversion_func)
        conv2 = Conversion("source1", "target2", mock_conversion_func)
        assert conv1 != conv2, "Instances with different targets should not be equal"

    def test_unequal_weights(self):
        """Test that instances with different weights are not equal."""
        conv1 = Conversion("source1", "target1", mock_conversion_func, weight=1)
        conv2 = Conversion("source1", "target1", mock_conversion_func, weight=0.6)
        assert conv1 != conv2, "Instances with different weights should not be equal"

    def test_comparison_with_different_type(self):
        """Test that a Conversion instance is not equal to an object of a different type."""
        conv = Conversion("source1", "target1", mock_conversion_func)
        assert (
            conv != "a string"
        ), "Comparison with an object of a different type should return False"


@pytest.mark.parametrize(
    "conversions, start, end, expected_path",
    [
        ([("a", "b", 1.00), ("b", "c", 1.00), ("a", "c", 0.77)], "a", "c", "a -> b -> c"),
        ([("a", "b", 1.00), ("b", "c", 1.00), ("a", "c", 0.78)], "a", "c", "a -> c"),
        (
            [("a", "b", 1.00), ("b", "c", 1.00), ("c", "d", 1.00), ("a", "d", 0.60)],
            "a",
            "d",
            "a -> b -> c -> d",
        ),
        (
            [("a", "b", 1.00), ("b", "c", 1.00), ("c", "d", 1.00), ("a", "d", 0.61)],
            "a",
            "d",
            "a -> d",
        ),
    ],
)
def test_conversion_weight_bias(conversions, start, end, expected_path):
    """Tests the effect of weight biases on determining the shortest path in a conversion graph."""
    conversions = [
        Conversion(src, dst, lambda x: x, weight, bias=0.25) for src, dst, weight in conversions
    ]
    graph = ConversionGraph(conversions=conversions)
    shortest_path = graph.shortest_path(start, end)
    assert shortest_path == expected_path



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining IonQ session and provider classes

"""
import os
from typing import Any, Optional

from qbraid_core.sessions import Session

from qbraid._caching import cached_method
from qbraid._version import __version__ as qbraid_version
from qbraid.programs import ExperimentType, ProgramSpec
from qbraid.programs.gate_model.ionq import (
    IONQ_NATIVE_GATES,
    IONQ_NATIVE_GATES_BASE,
    IONQ_NATIVE_GATES_FAMILY,
    IONQ_QIS_GATES,
)
from qbraid.runtime.exceptions import ResourceNotFoundError
from qbraid.runtime.noise import NoiseModelSet
from qbraid.runtime.profile import TargetProfile
from qbraid.runtime.provider import QuantumProvider
from qbraid.transpiler.conversions.qasm2.qasm2_to_ionq import qasm2_to_ionq
from qbraid.transpiler.conversions.qasm3.qasm3_to_ionq import qasm3_to_ionq

from .device import IonQDevice


class IonQSession(Session):
    """IonQ session class."""

    def __init__(self, api_key: Optional[str] = None):
        api_key = api_key or os.getenv("IONQ_API_KEY")
        if not api_key:
            raise ValueError(
                "An IonQ API key is required to initialize the session. "
                "Please provide it directly as an argument or set it via "
                "the IONQ_API_KEY environment variable."
            )

        super().__init__(
            base_url="https://api.ionq.co/v0.3",
            headers={"Content-Type": "application/json"},
            auth_headers={"Authorization": f"apiKey {api_key}"},
        )
        self.api_key = api_key
        self.add_user_agent(f"QbraidSDK/{qbraid_version}")

    def get_devices(self, **kwargs) -> dict[str, dict[str, Any]]:
        """Get all IonQ devices."""
        devices_list = self.get("/backends", **kwargs).json()
        devices_dict = {device["backend"]: device for device in devices_list}
        return devices_dict

    def get_device(self, device_id: str) -> dict[str, Any]:
        """Get a specific IonQ device."""
        devices = self.get_devices()
        try:
            return devices[device_id]
        except KeyError as err:
            raise ResourceNotFoundError(f"Device '{device_id}' not found.") from err

    def create_job(self, data: dict[str, Any]) -> dict[str, Any]:
        """Create a new job on the IonQ API."""
        return self.post("/jobs", data=data).json()

    def get_job(self, job_id: str) -> dict[str, Any]:
        """Get a specific IonQ job."""
        return self.get(f"/jobs/{job_id}").json()

    def cancel_job(self, job_id: str) -> dict[str, Any]:
        """Cancel a specific IonQ job."""
        return self.put(f"/jobs/{job_id}/status/cancel").json()


class IonQProvider(QuantumProvider):
    """IonQ provider class."""

    def __init__(self, api_key: Optional[str] = None):
        self.session = IonQSession(api_key)

    def _get_characterization(self, data: dict[str, Any]) -> Optional[dict[str, Any]]:
        """Return the characterization of the IonQ device."""
        characterization_url = data.get("characterization_url")
        if not characterization_url:
            return None
        characterization_endpoint = f"{self.session.base_url}{characterization_url}"
        return self.session.get(characterization_endpoint).json()

    @staticmethod
    def _get_basis_gates(device_id: str) -> list[str]:
        """Return the basis gates for the IonQ device."""
        if device_id == "simulator":
            native_gates = IONQ_NATIVE_GATES.copy()
        else:
            native_gates = next(
                (gates for key, gates in IONQ_NATIVE_GATES_FAMILY.items() if key in device_id),
                IONQ_NATIVE_GATES_BASE,
            )

        basis_gates = list(set(IONQ_QIS_GATES.copy() + native_gates))

        return basis_gates

    def _build_profile(self, data: dict[str, Any]) -> TargetProfile:
        """Build a profile for an IonQ device."""
        device_id = data.get("backend")
        simulator = device_id == "simulator"
        charact = self._get_characterization(data)
        basis_gates = self._get_basis_gates(device_id)
        noise_models = (
            NoiseModelSet.from_iterable(data.get("noise_models", [])) if simulator else None
        )

        return TargetProfile(
            device_id=device_id,
            simulator=simulator,
            experiment_type=ExperimentType.GATE_MODEL,
            num_qubits=data.get("qubits"),
            program_spec=[
                ProgramSpec(str, alias="qasm2", serialize=qasm2_to_ionq),
                ProgramSpec(str, alias="qasm3", serialize=qasm3_to_ionq),
            ],
            provider_name="IonQ",
            basis_gates=basis_gates,
            characterization=charact,
            noise_models=noise_models,
        )

    @cached_method
    def get_device(self, device_id: str) -> IonQDevice:
        """Get a specific IonQ device."""
        data = self.session.get_device(device_id)
        profile = self._build_profile(data)
        return IonQDevice(profile, self.session)

    @cached_method
    def get_devices(self, **kwargs) -> list[IonQDevice]:
        """Get a list of IonQ devices."""
        devices = self.session.get_devices(**kwargs)
        return [
            IonQDevice(self._build_profile(device), self.session) for device in devices.values()
        ]

    def __hash__(self):
        if not hasattr(self, "_hash"):
            object.__setattr__(self, "_hash", hash((self.session.api_key, self.session.base_url)))
        return self._hash  # pylint: disable=no-member



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for circuit drawer

# TODO: shared15 draw testcase, after fixing circuit decomposition problem from cirq

"""
import importlib.util

import pytest

from qbraid.programs import ProgramTypeError, load_program
from qbraid.visualization.draw_circuit import circuit_drawer


def is_package_installed(package_name: str) -> bool:
    """Check if a package is installed."""
    return importlib.util.find_spec(package_name) is not None


ALL_TARGETS = ["braket", "cirq", "qiskit", "pytket", "pyquil"]
AVAILABLE_TARGETS = [pkg for pkg in ALL_TARGETS if is_package_installed(pkg)] + ["qasm2"]


def test_draw_raises():
    """Test that non-supported package raises error"""
    with pytest.raises(ProgramTypeError):
        circuit_drawer("bad_input")


def test_draw_program_raises():
    """Test that non-supported package raises error"""
    with pytest.raises(ProgramTypeError):
        circuit_drawer(None)


@pytest.mark.skipif("qiskit" not in AVAILABLE_TARGETS, reason="qiskit not installed")
@pytest.mark.parametrize("bell_circuit", ["qiskit"], indirect=True)
def test_qiskit_draw(bell_circuit):
    """Test draw function standard output for qiskit bell circuit."""
    expected = """          ┌───┐
q_0: ─────┤ X ├
     ┌───┐└─┬─┘
q_1: ┤ H ├──■──
     └───┘     """
    qiskit_bell, _ = bell_circuit
    qprogram = load_program(qiskit_bell)
    qprogram.reverse_qubit_order()
    result = circuit_drawer(qprogram.program, output="text")
    assert str(result) == expected


@pytest.mark.skipif("braket" not in AVAILABLE_TARGETS, reason="braket not installed")
@pytest.mark.parametrize("bell_circuit", AVAILABLE_TARGETS, indirect=True)
def test_braket_bell_draw(capfd, bell_circuit):
    """Test using Amazon Braket to draw using ascii output is of the expected length."""
    circuit, _ = bell_circuit
    circuit_drawer(circuit, as_package="braket", output="ascii")

    _, err = capfd.readouterr()
    assert len(err) == 0


@pytest.mark.skipif("braket" not in AVAILABLE_TARGETS, reason="braket not installed")
@pytest.mark.parametrize("bell_circuit", ["braket"], indirect=True)
def test_braket_raises(bell_circuit):
    """Test that drawing braket circuit with non-supported output raises error"""
    braket_bell, _ = bell_circuit
    with pytest.raises(ValueError):
        circuit_drawer(braket_bell, output="bad_input")


@pytest.mark.skipif("cirq" not in AVAILABLE_TARGETS, reason="cirq not installed")
@pytest.mark.parametrize("bell_circuit", AVAILABLE_TARGETS, indirect=True)
def test_cirq_bell_text_draw(capfd, bell_circuit):
    """Test using Cirq to draw circuit using text output is of the expected length."""
    circuit, _ = bell_circuit
    circuit_drawer(circuit, as_package="cirq", output="text")

    _, err = capfd.readouterr()
    assert len(err) == 0


@pytest.mark.skipif("cirq" not in AVAILABLE_TARGETS, reason="cirq not installed")
@pytest.mark.parametrize("bell_circuit", ["cirq"], indirect=True)
def test_cirq_bell_svg_draw(bell_circuit):
    """Test drawing Cirq circuit using SVG source output"""
    cirq_bell, _ = bell_circuit
    svg_str = circuit_drawer(cirq_bell, output="svg_source")
    assert svg_str.startswith("<svg") and svg_str.endswith("</svg>")


@pytest.mark.skipif("cirq" not in AVAILABLE_TARGETS, reason="cirq not installed")
@pytest.mark.parametrize("bell_circuit", ["cirq"], indirect=True)
def test_cirq_raises(bell_circuit):
    """Test that drawing Cirq circuit with non-supported output raises error"""
    cirq_bell, _ = bell_circuit
    with pytest.raises(ValueError):
        circuit_drawer(cirq_bell, output="bad_input")


@pytest.mark.skipif("pyquil" not in AVAILABLE_TARGETS, reason="pyquil not installed")
@pytest.mark.parametrize("bell_circuit", AVAILABLE_TARGETS, indirect=True)
def test_pyquil_bell_draw(capfd, bell_circuit):
    """Test using pyquil to draw circuit using text output is of the expected length."""
    circuit, _ = bell_circuit
    circuit_drawer(circuit, as_package="pyquil", output="text")

    _, err = capfd.readouterr()
    assert len(err) == 0


@pytest.mark.skipif("pyquil" not in AVAILABLE_TARGETS, reason="pyquil not installed")
@pytest.mark.parametrize("bell_circuit", ["pyquil"], indirect=True)
def test_pyquil_raises(bell_circuit):
    """Test that drawing pyQuil program with non-supported output raises error"""
    pyquil_bell, _ = bell_circuit
    with pytest.raises(ValueError):
        circuit_drawer(pyquil_bell, output="bad_input")


@pytest.mark.skipif("pytket" not in AVAILABLE_TARGETS, reason="pytket not installed")
@pytest.mark.parametrize("bell_circuit", ["pytket"], indirect=True)
def test_pytket_draw(bell_circuit):
    """Test draw function html output for pytket bell circuit."""
    pytket_bell, _ = bell_circuit
    raw_html: str = circuit_drawer(pytket_bell, output="html")
    assert 2381 <= len(raw_html) <= 4468


@pytest.mark.skipif("pytket" not in AVAILABLE_TARGETS, reason="pytket not installed")
@pytest.mark.parametrize("bell_circuit", ["pytket"], indirect=True)
def test_pytket_raises(bell_circuit):
    """Test that drawing pytket circuit with non-supported output raises error"""
    pytket_bell, _ = bell_circuit
    with pytest.raises(ValueError):
        circuit_drawer(pytket_bell, output="bad_input")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for calculating unitary of quantum circuit/program

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from qbraid.programs import load_program

if TYPE_CHECKING:
    import qbraid


def match_global_phase(a: np.ndarray, b: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Matches the global phase of two numpy arrays.

    This function aligns the global phases of two matrices by applying a phase shift based
    on the position of the largest entry in one matrix. It computes and adjusts for the
    phase difference at this position, resulting in two phase-aligned matrices. The output,
    (a', b'), indicates that a' == b' is equivalent to a == b * exp(i * t) for some phase t.

    Args:
        a (np.ndarray): The first input matrix.
        b (np.ndarray): The second input matrix.

    Returns:
        tuple[np.ndarray, np.ndarray]: A tuple of the two matrices `(a', b')`, adjusted for
                                       global phase. If shapes of `a` and `b` do not match or
                                       either is empty, returns copies of the original matrices.
    """
    if a.shape != b.shape or a.size == 0:
        return np.copy(a), np.copy(b)

    k = max(np.ndindex(*a.shape), key=lambda t: abs(b[t]))

    def dephase(v):
        r = np.real(v)
        i = np.imag(v)

        if i == 0:
            return -1 if r < 0 else 1
        if r == 0:
            return 1j if i < 0 else -1j

        return np.exp(-1j * np.arctan2(i, r))

    return a * dephase(a[k]), b * dephase(b[k])


def assert_allclose_up_to_global_phase(a: np.ndarray, b: np.ndarray, atol: float, **kwargs) -> None:
    """
    Checks if two numpy arrays are equal up to a global phase, within
    a specified tolerance, i.e. if a ~= b * exp(i t) for some t.

    Args:
        a (np.ndarray): The first input array.
        b (np.ndarray): The second input array.
        atol (float): The absolute error tolerance.
        **kwargs: Additional keyword arguments to pass to `np.testing.assert_allclose`.

    Raises:
        AssertionError: The matrices aren't nearly equal up to global phase.

    """
    a, b = match_global_phase(a, b)
    np.testing.assert_allclose(actual=a, desired=b, atol=atol, **kwargs)


def circuits_allclose(  # pylint: disable=too-many-arguments
    circuit0: qbraid.programs.QPROGRAM,
    circuit1: qbraid.programs.QPROGRAM,
    index_contig: bool = False,
    allow_rev_qubits: bool = False,
    strict_gphase: bool = False,
    atol: float = 1e-7,
) -> bool:
    """Check if quantum program unitaries are equivalent.

    Args:
        circuit0 (:data:`~qbraid.programs.QPROGRAM`): First quantum program to compare
        circuit1 (:data:`~qbraid.programs.QPROGRAM`): Second quantum program to compare
        index_contig: If True, calculates circuit unitaries using contiguous qubit indexing.
        allow_rev_qubits: Whether to count identical circuits with reversed qubit ordering
            as equivalent.
        strict_gphase: If False, disregards global phase when verifying
            equivalence of the input circuit's unitaries.
        atol: Absolute tolerance parameter for np.allclose function.

    Returns:
        True if the input circuits pass unitary equality check

    """

    def unitary_equivalence_check(unitary0, unitary1, unitary_rev=None):
        if strict_gphase:
            return np.allclose(unitary0, unitary1) or (
                allow_rev_qubits and np.allclose(unitary0, unitary_rev)
            )
        try:
            assert_allclose_up_to_global_phase(unitary0, unitary1, atol=atol)
        except AssertionError:
            if allow_rev_qubits:
                try:
                    assert_allclose_up_to_global_phase(unitary0, unitary_rev, atol=atol)
                except AssertionError:
                    return False
            else:
                return False
        return True

    program0 = load_program(circuit0)
    program1 = load_program(circuit1)

    if index_contig:
        program0.remove_idle_qubits()
        program1.remove_idle_qubits()

    unitary0 = program0.unitary()
    unitary1 = program1.unitary()
    unitary_rev = program1.unitary_rev_qubits()

    return unitary_equivalence_check(unitary0, unitary1, unitary_rev)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for conversions between Cirq circuits and Qiskit circuits.

"""
import cirq
import numpy as np
import pyqasm
import pytest
import qiskit
from cirq import Circuit, LineQubit, ops, testing
from qiskit import QuantumCircuit
from qiskit.circuit.random import random_circuit

from qbraid.interface import circuits_allclose
from qbraid.programs import load_program
from qbraid.transpiler.conversions.cirq import cirq_to_qasm2
from qbraid.transpiler.conversions.qasm2 import qasm2_to_cirq
from qbraid.transpiler.conversions.qasm3 import qasm3_to_qiskit
from qbraid.transpiler.conversions.qiskit import qiskit_to_qasm2, qiskit_to_qasm3
from qbraid.transpiler.converter import transpile
from qbraid.transpiler.exceptions import ProgramConversionError

from ..cirq_utils import _equal


def test_bell_state_to_from_circuits():
    """Tests cirq.Circuit --> qiskit.QuantumCircuit --> cirq.Circuit
    with a Bell state circuit.
    """
    qreg = cirq.LineQubit.range(2)
    cirq_circuit = cirq.Circuit([cirq.ops.H.on(qreg[0]), cirq.ops.CNOT.on(qreg[0], qreg[1])])
    qiskit_circuit = transpile(cirq_circuit, "qiskit")  # Qiskit from Cirq
    circuit_cirq = transpile(qiskit_circuit, "cirq")  # Cirq from Qiskit
    assert np.allclose(cirq_circuit.unitary(), circuit_cirq.unitary())


def test_bell_state_to_qiskit():
    """Tests cirq.Circuit --> qiskit.QuantumCircuit --> cirq.Circuit
    with a Bell state circuit."""
    qreg = LineQubit.range(2)
    cirq_circuit = Circuit([ops.H.on(qreg[0]), ops.CNOT.on(qreg[0], qreg[1])])
    qiskit_circuit = transpile(cirq_circuit, "qiskit")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("num_qubits", [1, 2, 3, 4, 5])
def test_random_circuit_to_qiskit(num_qubits):
    """Tests converting random Cirq circuits to Qiskit circuits."""
    for _ in range(10):
        cirq_circuit = testing.random_circuit(
            qubits=num_qubits,
            n_moments=np.random.randint(1, 6),
            op_density=1,
            random_state=np.random.randint(1, 10),
        )
        qiskit_circuit = transpile(cirq_circuit, "qiskit")
        assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


def test_bell_state_to_from_qasm():
    """Tests cirq.Circuit --> QASM string --> cirq.Circuit
    with a Bell state circuit.
    """
    qreg = cirq.LineQubit.range(2)
    cirq_circuit = cirq.Circuit([cirq.ops.H.on(qreg[0]), cirq.ops.CNOT.on(qreg[0], qreg[1])])
    qasm = cirq_to_qasm2(cirq_circuit)  # Qasm from Cirq
    circuit_cirq = qasm2_to_cirq(qasm)
    assert np.allclose(cirq_circuit.unitary(), circuit_cirq.unitary())


def test_random_circuit_to_from_circuits():
    """Tests cirq.Circuit --> qiskit.QuantumCircuit --> cirq.Circuit
    with a random two-qubit circuit.
    """
    cirq_circuit = cirq.testing.random_circuit(
        qubits=2, n_moments=10, op_density=0.99, random_state=1
    )
    qiskit_circuit = transpile(cirq_circuit, "qiskit")
    circuit_cirq = transpile(qiskit_circuit, "cirq")
    assert np.allclose(cirq_circuit.unitary(), circuit_cirq.unitary())


def test_random_circuit_to_from_qasm():
    """Tests cirq.Circuit --> QASM string --> cirq.Circuit
    with a random one-qubit circuit.
    """
    circuit_0 = cirq.testing.random_circuit(qubits=2, n_moments=10, op_density=0.99, random_state=2)
    qasm = cirq_to_qasm2(circuit_0)
    circuit_1 = qasm2_to_cirq(qasm)
    u_0 = circuit_0.unitary()
    u_1 = circuit_1.unitary()
    assert cirq.equal_up_to_global_phase(u_0, u_1)


@pytest.mark.parametrize("as_qasm", (True, False))
def test_convert_with_barrier(as_qasm):
    """Tests converting a Qiskit circuit with a barrier to a Cirq circuit."""
    n = 5
    qiskit_circuit = qiskit.QuantumCircuit(qiskit.QuantumRegister(n))
    qiskit_circuit.barrier()

    if as_qasm:
        cirq_circuit = qasm2_to_cirq(qiskit_to_qasm2(qiskit_circuit))
    else:
        cirq_circuit = transpile(qiskit_circuit, "cirq")

    assert _equal(cirq_circuit, cirq.Circuit())


@pytest.mark.parametrize("as_qasm", (True, False))
def test_convert_with_multiple_barriers(as_qasm):
    """Tests converting a Qiskit circuit with barriers to a Cirq circuit."""
    n = 1
    num_ops = 10

    qreg = qiskit.QuantumRegister(n)
    qiskit_circuit = qiskit.QuantumCircuit(qreg)
    for _ in range(num_ops):
        qiskit_circuit.h(qreg)
        qiskit_circuit.barrier()

    if as_qasm:
        cirq_circuit = qasm2_to_cirq(qiskit_to_qasm2(qiskit_circuit))
    else:
        cirq_circuit = transpile(qiskit_circuit, "cirq")

    qbit = cirq.LineQubit(0)
    correct = cirq.Circuit(cirq.ops.H.on(qbit) for _ in range(num_ops))
    assert _equal(cirq_circuit, correct)
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


def test_bell_state_from_qiskit():
    """Tests qiskit.QuantumCircuit --> cirq.Circuit
    with a Bell state circuit.
    """
    qiskit_circuit = QuantumCircuit(2)
    qiskit_circuit.h(0)
    qiskit_circuit.cx(0, 1)
    cirq_circuit = transpile(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


def test_common_gates_from_qiskit():
    """Tests converting standard gates from Qiskit to Cirq."""
    qiskit_circuit = QuantumCircuit(4)
    qiskit_circuit.h([0, 1, 2, 3])
    qiskit_circuit.x([0, 1])
    qiskit_circuit.y(2)
    qiskit_circuit.z(3)
    qiskit_circuit.s(0)
    qiskit_circuit.sdg(1)
    qiskit_circuit.t(2)
    qiskit_circuit.tdg(3)
    qiskit_circuit.rx(np.pi / 4, 0)
    qiskit_circuit.ry(np.pi / 2, 1)
    qiskit_circuit.rz(3 * np.pi / 4, 2)
    qiskit_circuit.sx(0)
    qiskit_circuit.iswap(2, 3)
    qiskit_circuit.swap([0, 1], [2, 3])
    qiskit_circuit.cx(0, 1)
    cirq_circuit = transpile(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.skip(reason="Phase gates resulting in error")
def test_phase_gates_from_qiskit():
    """Tests converting standard gates from Qiskit to Cirq."""
    qiskit_circuit = QuantumCircuit(4)
    qiskit_circuit.p(np.pi / 8, 3)
    qiskit_circuit.sxdg(1)
    qiskit_circuit.cp(np.pi / 4, 2, 3)
    cirq_circuit = transpile(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubits", ([0, 1], [1, 0]))
def test_crz_gate_from_qiskit(qubits):
    """Tests converting controlled Rz gate from Qiskit to Cirq."""
    qiskit_circuit = QuantumCircuit(2)
    qiskit_circuit.crz(np.pi / 4, *qubits)
    cirq_circuit = transpile(qiskit_circuit, "cirq", require_native=True)
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


@pytest.mark.parametrize("qubits", ([0, 1], [1, 0]))
@pytest.mark.parametrize("theta", (0, np.pi, 2 * np.pi, np.pi / 2, np.pi / 4))
def test_rzz_gate_from_qiskit(qubits, theta):
    """Tests converting Rzz gate from Qiskit to Cirq."""
    qiskit_circuit = QuantumCircuit(2)
    qiskit_circuit.rzz(theta, *qubits)
    cirq_circuit = transpile(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


def test_iswap_gate_from_qiskit():
    """Tests converting iSwap gate from Qiskit to Cirq."""
    qiskit_circuit = QuantumCircuit(2)
    qiskit_circuit.h([0, 1])
    qiskit_circuit.iswap(0, 1)
    cirq_circuit = transpile(qiskit_circuit, "cirq")
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=True)


def test_qiskit_roundtrip():
    """Test converting qiskit gates that previously failed qiskit roundtrip test."""
    qiskit_circuit = QuantumCircuit(3)
    qiskit_circuit.ccz(0, 1, 2)
    qiskit_circuit.ecr(1, 2)
    qiskit_circuit.cs(2, 0)
    cirq_circuit = transpile(qiskit_circuit, "cirq", require_native=True)
    assert circuits_allclose(qiskit_circuit, cirq_circuit, strict_gphase=False)


def test_qiskit_roundtrip_noncontig():
    """Test converting gates that previously failed qiskit roundtrip test
    with non-contiguous qubits."""
    qiskit_circuit = QuantumCircuit(4)
    qiskit_circuit.ccz(0, 1, 2)
    qiskit_circuit.ecr(1, 2)
    qiskit_circuit.cs(2, 0)
    cirq_circuit = transpile(qiskit_circuit, "cirq", require_native=True)
    qprogram = load_program(cirq_circuit)
    qprogram.remove_idle_qubits()
    qiskit_contig = qprogram.program
    assert circuits_allclose(qiskit_contig, cirq_circuit, strict_gphase=False)


def test_100_random_qiskit():
    """Test converting 100 random qiskit circuits to cirq."""

    gates_to_skip = ["rxx", "rzz"]

    def _circuit_contains_invalid_gate(qiskit_circuit):
        for gate in gates_to_skip:
            if len(qiskit_circuit.get_instructions(gate)) > 0:
                return True
        return False

    def _generate_valid_qiskit_circuit():
        qiskit_circuit = random_circuit(5, 1, max_operands=4)
        while _circuit_contains_invalid_gate(qiskit_circuit):
            qiskit_circuit = random_circuit(5, 1, max_operands=4)
        return qiskit_circuit

    for _ in range(100):
        qiskit_circuit = _generate_valid_qiskit_circuit()

        cirq_circuit = transpile(qiskit_circuit, "cirq", require_native=True)
        assert circuits_allclose(
            qiskit_circuit, cirq_circuit, strict_gphase=False, index_contig=True
        )


def test_qiskit_to_from_qasm3():
    """Test converting qiskit circuit to/from OpenQASM 3.0 string"""
    circuit_in = QuantumCircuit(2)
    circuit_in.h(0)
    circuit_in.cx(0, 1)

    qasm3_str = qiskit_to_qasm3(circuit_in)
    circuit_out = qasm3_to_qiskit(qasm3_str)
    assert circuits_allclose(circuit_in, circuit_out, strict_gphase=True)


def test_raise_circuit_conversion_error():
    """Tests raising error for unsupported gates."""
    with pytest.raises(ProgramConversionError):
        probs = np.random.uniform(low=0, high=0.5)
        cirq_circuit = Circuit(ops.PhaseDampingChannel(probs).on(*LineQubit.range(1)))
        transpile(cirq_circuit, "qiskit")


def test_raise_qasm_error():
    """Test raising error for unsupported gates."""
    with pytest.raises(pyqasm.ValidationError):
        qiskit_circuit = QuantumCircuit(1)
        qiskit_circuit.delay(300, 0)
        qasm2 = qiskit_to_qasm2(qiskit_circuit)
        _ = qasm2_to_cirq(qasm2)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing granular type checking for quantum programs
that use Python's built-in types.

"""
from abc import ABCMeta, abstractmethod
from typing import Any, Optional, Type, TypeVar

from pyqasm.analyzer import Qasm3Analyzer
from pyqasm.exceptions import QasmParsingError

from .exceptions import QasmError
from .exceptions import ValidationError as ProgramValidationError

IonQDictType = TypeVar("IonQDictType", bound=dict)
QuboCoefficientsDictType = TypeVar("QuboCoefficientsDictType", bound=dict)


class QbraidMetaType(ABCMeta):
    """Abstract metaclass for custom program type checking based on built-in types."""

    @property
    @abstractmethod
    def __alias__(cls) -> str | None:
        """The program type alias to associate with the class."""

    @property
    @abstractmethod
    def __bound__(cls) -> Type[Any]:
        """The built-in type the class is wrapping."""


class QuboCoefficientsDictInstanceMeta(QbraidMetaType):
    """Metaclass for Qubo coefficients JSON type checking based on dict content."""

    @property
    def __alias__(cls) -> str:
        return "qubo"

    @property
    def __bound__(cls) -> Type[dict]:
        return dict

    def __instancecheck__(cls, instance: Any) -> bool:
        """Custom instance checks based on dict format."""
        if not instance or not isinstance(instance, dict):
            return False

        for key, value in instance.items():
            if not (
                isinstance(key, tuple) and len(key) == 2 and all(isinstance(k, str) for k in key)
            ):
                return False

            if not isinstance(value, (float, int)):
                return False

        return True


class QuboCoefficientsDict(metaclass=QuboCoefficientsDictInstanceMeta):
    """Marker class for dict that are valid Qubo coefficients format."""


class IonQDictInstanceMeta(QbraidMetaType):
    """Metaclass for IonQ JSON type checking based on dict content."""

    @property
    def __alias__(cls) -> str:
        return "ionq"

    @property
    def __bound__(cls) -> Type[dict]:
        return dict

    @staticmethod
    def _validate_field(single: Any, multiple: Any, field_name: str) -> None:
        """Helper method to validate single or multiple target/control fields."""
        if single is not None and multiple is not None:
            raise ProgramValidationError(
                f"Both {field_name} and {field_name}s are set; only one should be provided."
            )
        if single is not None and not isinstance(single, int):
            raise ProgramValidationError(f"Invalid {field_name}: {single}. Must be an integer.")
        if multiple is not None and not (
            isinstance(multiple, list) and all(isinstance(item, int) for item in multiple)
        ):
            raise ProgramValidationError(
                f"Invalid {field_name}s: {multiple}. Must be a list of integers."
            )

    def __instancecheck__(cls, instance: Any) -> bool:
        """Custom instance checks based on dict format."""
        if not isinstance(instance, dict):
            return False

        qubits = instance.get("qubits")
        circuit = instance.get("circuit")
        gateset = instance.get("gateset")
        ciruit_format = instance.get("format")

        if ciruit_format is not None and not isinstance(ciruit_format, str):
            return False

        if gateset is not None and not isinstance(gateset, str):
            return False

        if not isinstance(qubits, int) or not isinstance(circuit, list):
            return False

        for op in circuit:
            if not isinstance(op, dict):
                return False

            gate = op.get("gate")
            rotation = op.get("rotation")
            target, targets = op.get("target"), op.get("targets")
            control, controls = op.get("control"), op.get("controls")

            if not isinstance(gate, str):
                return False

            if rotation is not None and not isinstance(rotation, (int, float)):
                return False

            try:
                cls._validate_field(target, targets, "target")
                cls._validate_field(control, controls, "control")
            except ProgramValidationError:
                return False

        return True


class IonQDict(metaclass=IonQDictInstanceMeta):
    """Marker class for dict that are valid IonQ JSON formatted programs."""


class BaseQasmInstanceMeta(QbraidMetaType):
    """Metaclass for OpenQASM type checking based on string content.

    Attributes:
        version (int | str | None): The specific OpenQASM type to check for.
    """

    version: int | str | None = None
    extension: str | None = None

    @property
    def __alias__(cls) -> str | None:
        if isinstance(cls.version, int):
            return f"qasm{cls.version}{'_' + cls.extension if cls.extension else ''}"
        return None

    @property
    def __bound__(cls) -> Type[str]:
        return str

    def __instancecheck__(cls, instance: Any) -> bool:
        """Custom instance checks based on OpenQASM type.

        Args:
            instance: The object to check.

        Returns:
            bool: True if instance is a string matching the OpenQASM type, False otherwise.
        """
        if not isinstance(instance, str):
            return False
        try:
            return int(Qasm3Analyzer.extract_qasm_version(instance)) == cls.version
        except QasmParsingError:
            return False


class Qasm2StringMeta(BaseQasmInstanceMeta):
    """Metaclass for instances representing OpenQASM 2 strings."""

    version = 2


class Qasm3StringMeta(BaseQasmInstanceMeta):
    """Metaclass for instances representing OpenQASM 3 strings."""

    version = 3


class Qasm2KirinStringMeta(BaseQasmInstanceMeta):
    """Metaclass for instances representing OpenQASM 2 strings adapted for Kirin parser."""

    version = 2
    extension = "kirin"

    def __instancecheck__(cls, instance: Any) -> bool:
        """Custom instance checks for Kirin OpenQASM 2 type.

        Args:
            instance: The object to check.

        Returns:
            bool: True if instance is a string matching the
                Kirin OpenQASM 2 type, False otherwise.
        """
        return isinstance(instance, str) and instance.lstrip().startswith("KIRIN")


class Qasm2String(metaclass=Qasm2StringMeta):
    """Marker class for strings that are valid OpenQASM 2 programs."""


class Qasm3String(metaclass=Qasm3StringMeta):
    """Marker class for strings that are valid OpenQASM 3 programs."""


class Qasm2KirinString(metaclass=Qasm2KirinStringMeta):
    """Marker class for strings that are valid OpenQASM 2 programs."""


class QasmStringType(str):
    """Base class for OpenQASM string types, providing validation upon instantiation."""

    version: Optional[int] = None

    def __new__(cls, value):
        if not isinstance(value, str):
            raise TypeError("OpenQASM strings must be initialized with a string.")
        if not int(Qasm3Analyzer.extract_qasm_version(value)) == cls.version:
            raise ValueError(f"String does not conform to OpenQASM {cls.version} format.")
        return str.__new__(cls, value)


class Qasm2StringType(QasmStringType):
    """Specifically typed string for OpenQASM 2 formatted text."""

    version = 2


class Qasm3StringType(QasmStringType):
    """Specifically typed string for OpenQASM 3 formatted text."""

    version = 3


def get_qasm_type_alias(qasm: str) -> str:
    """
    Determines the type alias for an OpenQASM program based on its version.

    Args:
        qasm (str): The OpenQASM program string.

    Returns:
        str: The QASM version alias ('qasm2', 'qasm3', or 'qasm2_kirin').

    Raises:
        QasmError: If the string does not represent a valid OpenQASM program.
    """
    if isinstance(qasm, Qasm2String):
        return Qasm2String.__alias__
    if isinstance(qasm, Qasm3String):
        return Qasm3String.__alias__
    if isinstance(qasm, Qasm2KirinString):
        return Qasm2KirinString.__alias__
    raise QasmError("Could not determine the type alias: the OpenQASM program may be invalid.")


QBRAID_META_TYPES = {IonQDict, QuboCoefficientsDict}
BOUND_QBRAID_META_TYPES = {Qasm2String, Qasm3String, Qasm2KirinString}



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for runtime noise model classes.

"""

import threading

import pytest

from qbraid.runtime.noise import NoiseModel, NoiseModelSet


def test_noise_model_initialization():
    """Test NoiseModel initialization with valid inputs."""
    model = NoiseModel("ideal")
    assert model.name == "ideal"
    assert model.value == "ideal"
    assert model.description.startswith("The simulation is performed without any noise")

    model_with_desc = NoiseModel("custom", "Custom description")
    assert model_with_desc.name == "custom"
    assert model_with_desc.value == "custom"
    assert model_with_desc.description == "Custom description"


def test_noise_model_normalization():
    """Test that noise model names are normalized correctly."""
    model = NoiseModel("  Ideal Noise ")
    assert model.value == "ideal_noise"


def test_noise_model_invalid_name():
    """Test that invalid noise model names raise ValueError."""
    with pytest.raises(ValueError) as excinfo:
        NoiseModel("Invalid@Name")
    assert "Invalid noise model name" in str(excinfo.value)


def test_noise_model_long_description():
    """Test that overly long descriptions raise ValueError."""
    long_description = "a" * 121
    with pytest.raises(ValueError) as excinfo:
        NoiseModel("ideal", long_description)
    assert "Description must be 120 characters or fewer" in str(excinfo.value)


def test_noise_model_str():
    """Test __str__ method of NoiseModel."""
    model = NoiseModel("ideal")
    assert str(model) == "ideal"


def test_noise_model_repr():
    """Test __repr__ method of NoiseModel."""
    model = NoiseModel("ideal")
    assert repr(model) == "NoiseModel('ideal')"


def test_noise_model_equality():
    """Test equality checks for NoiseModel."""
    model1 = NoiseModel("ideal")
    model2 = NoiseModel("Ideal")
    model3 = NoiseModel("depolarizing")

    assert model1 == model2
    assert model1 != model3
    assert model1 == "ideal"
    assert model1 == "Ideal"
    assert model1 != "depolarizing"
    assert model1 != 123  # Should return NotImplemented


def test_noise_model_hash():
    """Test that NoiseModel instances are hashable."""
    model_set = {NoiseModel("ideal"), NoiseModel("Ideal"), NoiseModel("depolarizing")}
    assert len(model_set) == 2  # "ideal" and "depolarizing"


def test_noise_model_immutability():
    """Test that NoiseModel instances are immutable."""
    model = NoiseModel("ideal")
    with pytest.raises(AttributeError):
        model.value = "new_value"
    with pytest.raises(AttributeError):
        model.description = "new_description"


def test_noise_models_add_and_get():
    """Test adding and retrieving noise models in NoiseModels."""
    models = NoiseModelSet()
    models.add("ideal")
    models.add("custom", "Custom description")

    ideal_model = models.get("ideal")
    assert ideal_model.value == "ideal"

    custom_model = models.get("custom")
    assert custom_model.description == "Custom description"


def test_noise_models_add_existing_no_overwrite():
    """Test adding an existing noise model without overwrite."""
    models = NoiseModelSet()
    models.add("ideal")
    with pytest.raises(ValueError) as excinfo:
        models.add("Ideal", "Different description")
    assert "already exists with a different definition" in str(excinfo.value)


def test_noise_models_add_existing_with_overwrite():
    """Test adding an existing noise model with overwrite."""
    models = NoiseModelSet()
    models.add("ideal", "Original description")
    models.add("Ideal", "New description", overwrite=True)
    ideal_model = models.get("ideal")
    assert ideal_model.description == "New description"


def test_noise_models_remove():
    """Test removing a noise model."""
    models = NoiseModelSet()
    models.add("ideal")
    models.remove("ideal")
    assert "ideal" not in models


def test_noise_models_remove_nonexistent():
    """Test removing a nonexistent noise model."""
    models = NoiseModelSet()
    with pytest.raises(KeyError) as excinfo:
        models.remove("nonexistent")
    assert "Noise model 'nonexistent' not found" in str(excinfo.value)


def test_noise_models_discard():
    """Test discarding a noise model."""
    models = NoiseModelSet()
    models.add("ideal")
    models.discard("ideal")
    assert "ideal" not in models
    models.discard("ideal")  # Should not raise an error


def test_noise_models_clear():
    """Test clearing all noise models."""
    models = NoiseModelSet()
    models.add("ideal")
    models.add("custom")
    models.clear()
    assert len(models) == 0


def test_noise_models_update():
    """Test updating one NoiseModels instance with another."""
    models1 = NoiseModelSet()
    models1.add("ideal")
    models2 = NoiseModelSet()
    models2.add("custom")
    models1.update(models2)
    assert "ideal" in models1
    assert "custom" in models1


def test_noise_models_update_invalid_type():
    """Test updating with an invalid type."""
    models = NoiseModelSet()
    with pytest.raises(TypeError) as excinfo:
        models.update(["ideal"])
    assert "Can only update from a dict or another NoiseModelSet instance" in str(excinfo.value)


def test_noise_models_update_from_dict():
    """Test updating with a dict type."""
    models = NoiseModelSet()
    models.update({"key": "value"})
    assert models.get("key").description == "value"


def test_noise_models_iteration():
    """Test iteration over NoiseModels."""
    models = NoiseModelSet()
    models.add("ideal")
    models.add("custom")
    keys = list(models)
    assert sorted(keys) == ["custom", "ideal"]


def test_noise_models_len():
    """Test len() function on NoiseModels."""
    models = NoiseModelSet()
    assert len(models) == 0
    models.add("ideal")
    assert len(models) == 1


def test_noise_models_contains():
    """Test __contains__ method of NoiseModels."""
    models = NoiseModelSet()
    models.add("ideal")
    assert "ideal" in models
    assert "Ideal" in models
    assert "nonexistent" not in models


def test_noise_models_getitem():
    """Test __getitem__ method of NoiseModels."""
    models = NoiseModelSet()
    models.add("ideal")
    ideal_model = models["ideal"]
    assert ideal_model.value == "ideal"
    with pytest.raises(KeyError):
        _ = models["nonexistent"]


def test_noise_models_setitem():
    """Test __setitem__ method of NoiseModels."""
    models = NoiseModelSet()
    model = NoiseModel("ideal")
    models["ideal"] = model
    assert "ideal" in models
    with pytest.raises(ValueError) as excinfo:
        models["ideal"] = "not a NoiseModel"
    assert "Value must be a NoiseModel instance" in str(excinfo.value)
    with pytest.raises(ValueError) as excinfo:
        models["different_key"] = model
    assert "Key does not match the NoiseModel's normalized value" in str(excinfo.value)


def test_noise_models_delitem():
    """Test __delitem__ method of NoiseModels."""
    models = NoiseModelSet()
    models.add("ideal")
    del models["ideal"]
    assert "ideal" not in models
    with pytest.raises(KeyError):
        del models["ideal"]


def test_noise_models_values():
    """Test values() method of NoiseModels."""
    models = NoiseModelSet()
    models.add("ideal")
    models.add("custom")
    values = list(models.values())
    assert len(values) == 2
    assert all(isinstance(model, NoiseModel) for model in values)


def test_noise_models_items():
    """Test items() method of NoiseModels."""
    models = NoiseModelSet()
    models.add("ideal")
    models.add("custom")
    items = list(models.items())
    assert len(items) == 2
    for key, model in items:
        assert key == model.value


def test_noise_models_repr():
    """Test __repr__ method of NoiseModels."""
    models = NoiseModelSet()
    models.add("ideal")
    models.add("custom")

    repr_str = repr(models)

    assert repr_str.startswith("NoiseModelSet([")
    assert "'ideal'" in repr_str
    assert "'custom'" in repr_str
    assert repr_str.endswith("])")


def test_noise_model_invalid_equality():
    """Test __eq__ with unsupported type."""
    model = NoiseModel("ideal")
    assert model != 123


def test_noise_models_thread_safety():
    """Test that NoiseModels is thread-safe."""
    models = NoiseModelSet()

    def add_model(name):
        models.add(name)

    threads = []
    for i in range(10):
        t = threading.Thread(target=add_model, args=(f"model_{i}",))
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

    assert len(models) == 10


def test_noise_models_invalid_key_in_setitem():
    """Test setting an item with a key that doesn't match the model's value."""
    models = NoiseModelSet()
    model = NoiseModel("ideal")
    with pytest.raises(ValueError) as excinfo:
        models["different_key"] = model
    assert "Key does not match the NoiseModel's normalized value" in str(excinfo.value)


def test_noise_models_pop():
    """Test popping a noise model."""
    models = NoiseModelSet()
    models.add("ideal")
    model = models._models.pop("ideal")
    assert model.value == "ideal"
    assert "ideal" not in models


def test_noise_model_normalize_staticmethod():
    """Test the _normalize static method."""
    assert NoiseModel._normalize("  Ideal Noise ") == "ideal_noise"


def test_noise_model_validate_classmethod():
    """Test the _validate class method."""
    # Valid case
    NoiseModel._validate("ideal_noise", "Valid description")
    # Invalid name
    with pytest.raises(ValueError):
        NoiseModel._validate("invalid@name")
    # Description too long
    long_description = "a" * 121
    with pytest.raises(ValueError):
        NoiseModel._validate("ideal_noise", long_description)


def test_noise_models_add_invalid_model():
    """Test adding a model with invalid name."""
    models = NoiseModelSet()
    with pytest.raises(ValueError):
        models.add("Invalid@Name")


def test_noise_models_get_nonexistent():
    """Test getting a nonexistent noise model."""
    models = NoiseModelSet()
    assert models.get("nonexistent") is None


def test_noise_models_discard_nonexistent():
    """Test discarding a nonexistent noise model."""
    models = NoiseModelSet()
    models.discard("nonexistent")  # Should not raise an error


def test_noise_models_update_with_overlapping_keys():
    """Test updating with overlapping keys."""
    models1 = NoiseModelSet()
    models1.add("ideal")
    models2 = NoiseModelSet()
    models2.add("ideal", "New description")
    models1.update(models2)
    ideal_model = models1.get("ideal")
    assert ideal_model.description == "New description"


def test_noise_models_invalid_type_in_contains():
    """Test __contains__ with invalid key type."""
    models = NoiseModelSet()
    assert 123 not in models


def test_noise_models_lock():
    """Test that the lock is working (indirectly)."""
    models = NoiseModelSet()
    models.add("ideal")

    def remove_model():
        models.remove("ideal")

    with models._lock:
        t = threading.Thread(target=remove_model)
        t.start()
        t.join(timeout=1)
        assert t.is_alive()  # Thread should be blocked


def test_noise_models_from_dict():
    """Test creating NoiseModelSet from a dictionary."""
    data = {"ideal": "Ideal noise model", "custom": "Custom noise model"}
    models = NoiseModelSet.from_dict(data)
    assert len(models) == 2
    assert models.get("ideal").description == "Ideal noise model"
    assert models.get("custom").description == "Custom noise model"



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing top-level qbraid job loader functionality
utilizing entrypoints via ``importlib``

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Literal, overload

from qbraid._entrypoints import get_entrypoints, load_entrypoint
from qbraid.exceptions import QbraidError

if TYPE_CHECKING:
    from qbraid.runtime import QuantumJob, QuantumProvider
    from qbraid.runtime.aws import BraketProvider, BraketQuantumTask
    from qbraid.runtime.azure import AzureQuantumJob, AzureQuantumProvider
    from qbraid.runtime.ibm import QiskitJob, QiskitRuntimeProvider
    from qbraid.runtime.ionq import IonQJob, IonQProvider
    from qbraid.runtime.native import QbraidJob, QbraidProvider
    from qbraid.runtime.oqc import OQCJob, OQCProvider


class JobLoaderError(QbraidError):
    """Raised when an error occurs while loading a quantum job."""


class ProviderLoaderError(QbraidError):
    """Raised when an error occurs while loading a quantum provider."""


@overload
def load_job(job_id: str, provider: Literal["native", "qbraid"], **kwargs) -> QbraidJob: ...


@overload
def load_job(job_id: str, provider: Literal["aws", "braket"], **kwargs) -> BraketQuantumTask: ...


@overload
def load_job(job_id: str, provider: Literal["ibm", "qiskit"], **kwargs) -> QiskitJob: ...


@overload
def load_job(job_id: str, provider: Literal["azure"], **kwargs) -> AzureQuantumJob: ...


@overload
def load_job(job_id: str, provider: Literal["ionq"], **kwargs) -> IonQJob: ...


@overload
def load_job(job_id: str, provider: Literal["oqc"], **kwargs) -> OQCJob: ...


@overload
def load_job(job_id: str, provider: str, **kwargs) -> QuantumJob: ...


def load_job(job_id: str, provider: str = "qbraid", **kwargs) -> QuantumJob:
    """Load a quantum job object from a supported provider.

    Args:
        job_id: The provider-specific job identifier.
        provider: The name of the provider module within in the ``qbraid.runtime`` package.

    Returns:
        QuantumJob: A quantum job object of the inferred subclass.

    Raises:
        JobLoaderError: If the job cannot be loaded.
    """
    provider_aliases = {"native": "qbraid", "qiskit": "ibm", "braket": "aws"}

    provider_module = provider_aliases.get(provider.lower(), provider).lower()

    try:
        job_class = load_entrypoint("jobs", provider_module)
    except Exception as err:
        raise JobLoaderError(
            f"Error loading QuantumJob sub-class for provider '{provider}'."
        ) from err

    job_instance = job_class(job_id, **kwargs)

    return job_instance


def get_providers() -> list[str]:
    """Return a list of supported quantum job providers.

    Returns:
        list[str]: A list of supported quantum job providers.
    """
    entry_points = get_entrypoints("providers")
    return sorted(entry_points.keys())


@overload
def load_provider(provider_name: Literal["native", "qbraid"], **kwargs) -> QbraidProvider: ...


@overload
def load_provider(provider_name: Literal["aws", "braket"], **kwargs) -> BraketProvider: ...


@overload
def load_provider(provider_name: Literal["ibm", "qiskit"], **kwargs) -> QiskitRuntimeProvider: ...


@overload
def load_provider(provider_name: Literal["azure"], **kwargs) -> AzureQuantumProvider: ...


@overload
def load_provider(provider_name: Literal["ionq"], **kwargs) -> IonQProvider: ...


@overload
def load_provider(provider_name: Literal["oqc"], **kwargs) -> OQCProvider: ...


@overload
def load_provider(provider_name: str, **kwargs) -> QuantumProvider: ...


def load_provider(provider_name: str = "qbraid", **kwargs) -> QuantumProvider:
    """Load a quantum provider object from a supported qBraid runtime module.

    Args:
        provider_name: The name of the provider module within in the
            ``qbraid.runtime`` package. Defaults to "qbraid".

    Returns:
        QuantumProvider: A quantum provider object of the inferred subclass.

    Raises:
        ProviderLoaderError: If the provider subclass cannot be loaded.
    """
    provider_aliases = {"native": "qbraid", "qiskit": "ibm", "braket": "aws"}

    provider_module = provider_aliases.get(provider_name, provider_name).lower()

    try:
        provider_class = load_entrypoint("providers", provider_module)
    except Exception as err:
        raise ProviderLoaderError(
            f"Error loading QuantumProvider sub-class for provider '{provider_name}'."
        ) from err

    provider_instance = provider_class(**kwargs)

    return provider_instance
